"""
ZCAM Colour Appearance Model
============================

Defines the *ZCAM* colour appearance model objects:

-   :class:`colour.appearance.InductionFactors_ZCAM`
-   :attr:`colour.VIEWING_CONDITIONS_ZCAM`
-   :class:`colour.CAM_Specification_ZCAM`
-   :func:`colour.XYZ_to_ZCAM`
-   :func:`colour.ZCAM_to_XYZ`

References
----------
-   :cite:`Safdar2018` : Safdar, M., Hardeberg, J. Y., Kim, Y. J., & Luo, M. R.
    (2018). A Colour Appearance Model based on J z a z b z Colour Space. Color
    and Imaging Conference, 2018(1), 96-101.
    doi:10.2352/ISSN.2169-2629.2018.26.96
-   :cite:`Safdar2021` : Safdar, M., Hardeberg, J. Y., & Ronnier Luo, M.
    (2021). ZCAM, a colour appearance model based on a high dynamic range
    uniform colour space. Optics Express, 29(4), 6036. doi:10.1364/OE.413659
-   :cite:`Zhai2018` : Zhai, Q., & Luo, M. R. (2018). Study of chromatic
    adaptation via neutral white matches on different viewing media. Optics
    Express, 26(6), 7724. doi:10.1364/OE.26.007724
"""

from __future__ import annotations

import numpy as np
from collections import namedtuple
from dataclasses import astuple, dataclass, field

from colour.adaptation import chromatic_adaptation_Zhai2018
from colour.appearance.ciecam02 import (
    VIEWING_CONDITIONS_CIECAM02,
    degree_of_adaptation,
    hue_angle,
)
from colour.algebra import spow
from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import (
    ArrayLike,
    Boolean,
    Dict,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    NDArray,
    Optional,
)
from colour.models import Izazbz_to_XYZ, XYZ_to_Izazbz, xy_to_XYZ
from colour.utilities import (
    CaseInsensitiveMapping,
    MixinDataclassArithmetic,
    as_float,
    as_float_array,
    as_int_array,
    domain_range_scale,
    from_range_1,
    from_range_degrees,
    has_only_nan,
    ones,
    to_domain_1,
    to_domain_degrees,
    tsplit,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "InductionFactors_ZCAM",
    "VIEWING_CONDITIONS_ZCAM",
    "CAM_Specification_ZCAM",
    "XYZ_to_ZCAM",
    "ZCAM_to_XYZ",
]


class InductionFactors_ZCAM(
    namedtuple("InductionFactors_ZCAM", ("F_s", "F", "c", "N_c"))
):
    """
    *ZCAM* colour appearance model induction factors.

    Parameters
    ----------
    F_s
        Surround impact :math:`F_s`.
    F
        Maximum degree of adaptation :math:`F`.
    c
        Exponential non-linearity :math:`c`.
    N_c
        Chromatic induction factor :math:`N_c`.

    Notes
    -----
    -   The *ZCAM* colour appearance model induction factors are inherited from
        the *CIECAM02* colour appearance model.

    References
    ----------
    :cite:`Safdar2021`
    """


VIEWING_CONDITIONS_ZCAM: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Average": InductionFactors_ZCAM(
            0.69, *VIEWING_CONDITIONS_CIECAM02["Average"]
        ),
        "Dim": InductionFactors_ZCAM(
            0.59, *VIEWING_CONDITIONS_CIECAM02["Dim"]
        ),
        "Dark": InductionFactors_ZCAM(
            0.525, *VIEWING_CONDITIONS_CIECAM02["Dark"]
        ),
    }
)
VIEWING_CONDITIONS_ZCAM.__doc__ = """
Reference *ZCAM* colour appearance model viewing conditions.

References
----------
:cite:`Safdar2021`
"""

HUE_DATA_FOR_HUE_QUADRATURE: Dict = {
    "h_i": np.array([33.44, 89.29, 146.30, 238.36, 393.44]),
    "e_i": np.array([0.68, 0.64, 1.52, 0.77, 0.68]),
    "H_i": np.array([0.0, 100.0, 200.0, 300.0, 400.0]),
}


@dataclass
class CAM_ReferenceSpecification_ZCAM(MixinDataclassArithmetic):
    """
    Define the *ZCAM* colour appearance model reference specification.

    This specification has field names consistent with :cite:`Safdar2021`
    reference.

    Parameters
    ----------
    J_z
        Correlate of *Lightness* :math:`J_z`.
    C_z
        Correlate of *chroma* :math:`C_z`.
    h_z
        *Hue* angle :math:`h_z` in degrees.
    S_z
        Correlate of *saturation* :math:`S_z`.
    Q_z
        Correlate of *brightness* :math:`Q_z`.
    M_z
        Correlate of *colourfulness* :math:`M_z`.
    H
        *Hue* :math:`h` quadrature :math:`H`.
    H_z
        *Hue* :math:`h` composition :math:`H_z`.
    V_z
        Correlate of *vividness* :math:`V_z`.
    K_z
        Correlate of *blackness* :math:`K_z`.
    W_z
        Correlate of *whiteness* :math:`W_z`.

    References
    ----------
    :cite:`Safdar2021`
    """

    J_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    C_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    h_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    S_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    Q_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    M_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    H: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    H_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    V_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    K_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    W_z: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)


@dataclass
class CAM_Specification_ZCAM(MixinDataclassArithmetic):
    """
    Define the *ZCAM* colour appearance model specification.

    Parameters
    ----------
    J
        *Lightness* :math:`J` is the "brightness of an area (:math:`Q`) judged
        relative to the brightness of a similarly illuminated area that appears
        to be white or highly transmitting (:math:`Q_w`)", i.e.,
        :math:`J = (Q/Q_w)`. It is a visual scale with two well defined levels
        i.e., zero and 100 for a pure black and a reference white,
        respectively. Note that in HDR visual field, samples could have a
        higher luminance than that of the reference white, so the lightness
        could be over 100. Subscripts :math:`s` and :math:`w` are used to
        annotate the sample and the reference white, respectively.
    C
        *Chroma* :math:`C` is "colourfulness of an area (:math:`M`) judged as
        a proportion of the brightness of a similarly illuminated area that
        appears white or highly transmitting (:math:`Q_w`)", i.e.,
        :math:`C = (M/Q_w)`. It is an open-end scale with origin as a colour
        in the neutral axis. It can be estimated as the magnitude of the
        chromatic difference between the test colour and a neutral colour
        having the lightness same as the test colour.
    h
        *Hue* angle :math:`h` is a scale ranged from :math:`0^{\\circ}` to
        :math:`360^{\\circ}` with the hues following rainbow sequence. The same
        distance between pairs of hues in a constant lightness and chroma shows
        the same perceived colour difference.
    s
        *Saturation* :math:`s` is the "colourfulness (:math:`M`) of an area
        judged in proportion to its brightness (:math:`Q`)", i.e.,
        :math:`s = (M/Q)`. It can also be defined as the chroma of an area
        judged in proportion to its lightness, i.e., :math:`s = (C/J)`. It is
        an open-end scale with all neutral colours to have saturation of zero.
        For example, the red bricks in a building would exhibit different
        colours when illuminated by daylight. Those (directly) under daylight
        will appear to be bright and colourful, and those under shadow will
        appear darker and less colourful. However, the two areas have the same
        saturation.
    Q
        *Brightness* :math:`Q` is an "attribute of a visual perception
        according to which an area appears to emit, or reflect, more or less
        light". It is an open-end scale with origin as pure black or complete
        darkness. It is an absolute scale according to the illumination
        condition i.e., an increase of brightness of an object when the
        illuminance of light is increased. This is a visual phenomenon known as
        Stevens effect.
    M
        *Colourfulness* :math:`M` is an "attribute of a visual perception
        according to which the perceived colour of an area appears to be more
        or less chromatic". It is an open-end scale with origin as a neutral
        colour i.e., appearance of no hue. It is an absolute scale according to
        the illumination condition i.e., an increase of colourfulness of an
        object when the illuminance of light is increased. This is a visual
        phenomenon known as Hunt effect.
    H
        *Hue* :math:`h` quadrature :math:`H_C` is an "attribute of a visual
        perception according to which an area appears to be similar to one of
        the colours: red, yellow, green, and blue, or to a combination of
        adjacent pairs of these colours considered in a closed ring". It has
        a 0-400 scale, i.e., hue quadrature of 0, 100, 200, 300, and 400
        range from unitary red to, yellow, green, blue, and back to red,
        respectively. For example, a cyan colour consists of 50% green and
        50% blue, corresponding to a hue quadrature of 250.
    HC
        *Hue* :math:`h` composition :math:`H^C` used to define the hue
        appearance of a sample. Note that hue circles formed by the equal hue
        angle and equal hue composition appear to be quite different.
    V
        *Vividness* :math:`V` is an "attribute of colour used to indicate the
        degree of departure of the  colour (of stimulus) from a neutral black
        colour", i.e., :math:`V = \\sqrt{J^2 + C^2}`. It is an open-end scale
        with origin at pure black. This reflects the visual phenomena of an
        object illuminated by a light to increase both the lightness and the
        chroma.
    K
        *Blackness* :math:`K` is a visual attribute according to which an area
        appears to contain more or less black content. It is a scale in the
        Natural Colour System (NCS) and can also be defined in resemblance to a
        pure black. It is an open-end scale with 100 as pure black (luminance
        of 0 :math:`cd/m^2`), i.e.,
        :math:`K = (100 - \\sqrt{J^2 + C^2} = (100 - V)`. The visual effect can
        be illustrated by mixing a black to a colour pigment. The more black
        pigment is added, the higher blackness will be. A blacker colour will
        have less lightness and/or chroma than a less black colour.
    W
        *Whiteness* :math:`W` is a visual attribute according to which an area
        appears to contain more or less white content. It is a scale of the NCS
        and can also be defined in resemblance to a pure white. It is an
        open-end scale with 100 as reference white, i.e.,
        :math:`W = (100 - \\sqrt{(100 - J)^2 + C^2} = (100 - D)`. The visual
        effect can be illustrated by mixing a white to a colour pigment. The
        more white pigment is added, the higher whiteness will be. A whiter
        colour will have a lower chroma and higher lightness than the less
        white colour.

    References
    ----------
    :cite:`Safdar2021`
    """

    J: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    C: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    h: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    s: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    Q: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    M: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    H: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    HC: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    V: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    K: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)
    W: Optional[FloatingOrNDArray] = field(default_factory=lambda: None)


TVS_D65: NDArray = xy_to_XYZ(
    CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"]["D65"]
)


def XYZ_to_ZCAM(
    XYZ: ArrayLike,
    XYZ_w: ArrayLike,
    L_A: FloatingOrArrayLike,
    Y_b: FloatingOrArrayLike,
    surround: InductionFactors_ZCAM = VIEWING_CONDITIONS_ZCAM["Average"],
    discount_illuminant: Boolean = False,
) -> CAM_Specification_ZCAM:
    """
    Compute the *ZCAM* colour appearance model correlates from given *CIE XYZ*
    tristimulus values.

    Parameters
    ----------
    XYZ
        Absolute *CIE XYZ* tristimulus values of test sample / stimulus.
    XYZ_w
        Absolute *CIE XYZ* tristimulus values of the white under reference
        illuminant.
    L_A
        Test adapting field *luminance* :math:`L_A` in :math:`cd/m^2` such as
        :math:`L_A = L_w * Y_b / 100` (where :math:`L_w` is luminance of the
        reference white and :math:`Y_b` is the background luminance factor).
    Y_b
        Luminous factor of background :math:`Y_b` such as
        :math:`Y_b = 100 * L_b / L_w` where :math:`L_w` is the luminance of the
        light source and :math:`L_b` is the luminance of the background. For
        viewing images, :math:`Y_b` can be the average :math:`Y` value for the
        pixels in the entire image, or frequently, a :math:`Y` value of 20,
        approximate an :math:`L^*` of 50 is used.
    surround
        Surround viewing conditions induction factors.
    discount_illuminant
        Truth value indicating if the illuminant should be discounted.

    Returns
    -------
    :class:`colour.CAM_Specification_ZCAM`
       *ZCAM* colour appearance model specification.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   *Safdar, Hardeberg and Luo (2021)* does not specify how the chromatic
        adaptation to *CIE Standard Illuminant D65* in *Step 0* should be
        performed. A one-step *Von Kries* chromatic adaptation transform is not
        symmetrical or transitive when a degree of adaptation is involved.
        *Safdar, Hardeberg and Luo (2018)* uses *Zhai and Luo (2018)* two-steps
        chromatic adaptation transform, thus it seems sensible to adopt this
        transform for the *ZCAM* colour appearance model until more information
        is available. It is worth noting that a one-step *Von Kries* chromatic
        adaptation transform with support for degree of adaptation produces
        values closer to the supplemental document compared to the
        *Zhai and Luo (2018)* two-steps chromatic adaptation transform but then
        the *ZCAM* colour appearance model does not round-trip properly.
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [UN]                  | [UN]          |
    +------------+-----------------------+---------------+
    | ``XYZ_tw`` | [UN]                  | [UN]          |
    +------------+-----------------------+---------------+
    | ``XYZ_rw`` | [UN]                  | [UN]          |
    +------------+-----------------------+---------------+

    +-------------------------------+-----------------------+---------------+
    | **Range**                     | **Scale - Reference** | **Scale - 1** |
    +===============================+=======================+===============+
    | ``CAM_Specification_ZCAM.J``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.C``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.h``  | [0, 360]              | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.s``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.Q``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.M``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.H``  | [0, 400]              | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.HC`` | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.V``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.K``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.H``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+

    References
    ----------
    :cite:`Safdar2018`, :cite:`Safdar2021`, :cite:`Zhai2018`

    Examples
    --------
    >>> XYZ = np.array([185, 206, 163])
    >>> XYZ_w = np.array([256, 264, 202])
    >>> L_A = 264
    >>> Y_b = 100
    >>> surround = VIEWING_CONDITIONS_ZCAM['Average']
    >>> XYZ_to_ZCAM(XYZ, XYZ_w, L_A, Y_b, surround)
    ... # doctest: +ELLIPSIS
    CAM_Specification_ZCAM(J=92.2504437..., C=3.0216926..., h=196.3245737..., \
s=19.1319556..., Q=321.3408463..., M=10.5256217..., H=237.6114442..., \
HC=None, V=34.7006776..., K=25.8835968..., W=91.6821728...)
    """

    XYZ = to_domain_1(XYZ)
    XYZ_w = to_domain_1(XYZ_w)
    _X_w, Y_w, _Z_w = tsplit(XYZ_w)
    L_A = as_float_array(L_A)
    Y_b = as_float_array(Y_b)

    F_s, _F, _c, _N_c = surround

    # Step 0 (Forward) - Chromatic adaptation from reference illuminant to
    # "CIE Standard Illuminant D65" illuminant using "CAT02".
    # Computing degree of adaptation :math:`D`.
    D = (
        degree_of_adaptation(surround.F, L_A)
        if not discount_illuminant
        else ones(L_A.shape)
    )

    XYZ_D65 = chromatic_adaptation_Zhai2018(
        XYZ, XYZ_w, TVS_D65, D, D, transform="CAT02"
    )

    # Step 1 (Forward) - Computing factors related with viewing conditions and
    # independent of the test stimulus.
    # Background factor :math:`F_b`
    F_b = np.sqrt(Y_b / Y_w)
    # Luminance level adaptation factor :math:`F_L`
    F_L = 0.171 * spow(L_A, 1 / 3) * (1 - np.exp(-48 / 9 * L_A))

    # Step 2 (Forward) - Computing achromatic response (:math:`I_z` and
    # :math:`I_{z,w}`), redness-greenness (:math:`a_z` and :math:`a_{z,w}`),
    # and yellowness-blueness (:math:`b_z`, :math:`b_{z,w}`).
    with domain_range_scale("ignore"):
        I_z, a_z, b_z = tsplit(XYZ_to_Izazbz(XYZ_D65, method="Safdar 2021"))
        I_z_w, _a_z_w, b_z_w = tsplit(
            XYZ_to_Izazbz(XYZ_w, method="Safdar 2021")
        )

    # Step 3 (Forward) - Computing hue angle :math:`h_z`
    h_z = hue_angle(a_z, b_z)

    # Step 4 (Forward) - Computing hue quadrature :math:`H`.
    H = hue_quadrature(h_z)

    # Computing eccentricity factor :math:`e_z`.
    e_z = 1.015 + np.cos(np.radians(89.038 + h_z % 360))

    # Step 5 (Forward) - Computing brightness :math:`Q_z`,
    # lightness :math:`J_z`, colourfulness :math`M_z`, and chroma :math:`C_z`
    Q_z_p = (1.6 * F_s) / F_b**0.12
    Q_z_m = F_s**2.2 * F_b**0.5 * spow(F_L, 0.2)
    Q_z = 2700 * spow(I_z, Q_z_p) * Q_z_m
    Q_z_w = 2700 * spow(I_z_w, Q_z_p) * Q_z_m

    J_z = 100 * (Q_z / Q_z_w)

    M_z = (
        100
        * (a_z**2 + b_z**2) ** 0.37
        * (
            (spow(e_z, 0.068) * spow(F_L, 0.2))
            / (F_b**0.1 * spow(I_z_w, 0.78))
        )
    )

    C_z = 100 * (M_z / Q_z_w)

    # Step 6 (Forward) - Computing saturation :math:`S_z`,
    # vividness :math:`V_z`, blackness :math:`K_z`, and whiteness :math:`W_z`.
    S_z = 100 * spow(F_L, 0.6) * np.sqrt(M_z / Q_z)

    V_z = np.sqrt((J_z - 58) ** 2 + 3.4 * C_z**2)

    K_z = 100 - 0.8 * np.sqrt(J_z**2 + 8 * C_z**2)

    W_z = 100 - np.sqrt((100 - J_z) ** 2 + C_z**2)

    return CAM_Specification_ZCAM(
        as_float(from_range_1(J_z)),
        as_float(from_range_1(C_z)),
        as_float(from_range_degrees(h_z)),
        as_float(from_range_1(S_z)),
        as_float(from_range_1(Q_z)),
        as_float(from_range_1(M_z)),
        as_float(from_range_degrees(H, 400)),
        None,
        as_float(from_range_1(V_z)),
        as_float(from_range_1(K_z)),
        as_float(from_range_1(W_z)),
    )


def ZCAM_to_XYZ(
    specification: CAM_Specification_ZCAM,
    XYZ_w: ArrayLike,
    L_A: FloatingOrArrayLike,
    Y_b: FloatingOrArrayLike,
    surround: InductionFactors_ZCAM = VIEWING_CONDITIONS_ZCAM["Average"],
    discount_illuminant: Boolean = False,
) -> NDArray:
    """
    Convert from *ZCAM* specification to *CIE XYZ* tristimulus values.

    Parameters
    ----------
    specification
         *ZCAM* colour appearance model specification.
         Correlate of *Lightness* :math:`J`, correlate of *chroma* :math:`C` or
         correlate of *colourfulness* :math:`M` and *hue* angle :math:`h` in
         degrees must be specified, e.g. :math:`JCh` or :math:`JMh`.
    XYZ_w
        Absolute *CIE XYZ* tristimulus values of the white under reference
        illuminant.
    L_A
        Test adapting field *luminance* :math:`L_A` in :math:`cd/m^2` such as
        :math:`L_A = L_w * Y_b / 100` (where :math:`L_w` is luminance of the
        reference white and :math:`Y_b` is the background luminance factor).
    Y_b
        Luminous factor of background :math:`Y_b` such as
        :math:`Y_b = 100 x L_b / L_w` where :math:`L_w` is the luminance of the
        light source and :math:`L_b` is the luminance of the background. For
        viewing images, :math:`Y_b` can be the average :math:`Y` value for the
        pixels in the entire image, or frequently, a :math:`Y` value of 20,
        approximate an :math:`L^*` of 50 is used.
    surround
        Surround viewing conditions induction factors.
    discount_illuminant
        Truth value indicating if the illuminant should be discounted.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* tristimulus values.

    Raises
    ------
    ValueError
        If neither *C* or *M* correlates have been defined in the
        ``CAM_Specification_ZCAM`` argument.

    Warnings
    --------
    The underlying *SMPTE ST 2084:2014* transfer function is an absolute
    transfer function.

    Notes
    -----
    -   *Safdar, Hardeberg and Luo (2021)* does not specify how the chromatic
        adaptation to *CIE Standard Illuminant D65* in *Step 0* should be
        performed. A one-step *Von Kries* chromatic adaptation transform is not
        symetrical or transitive when a degree of adptation is involved.
        *Safdar, Hardeberg and Luo (2018)* uses *Zhai and Luo (2018)* two-steps
        chromatic adaptation transform, thus it seems sensible to adopt this
        transform for the *ZCAM* colour appearance model until more information
        is available. It is worth noting that a one-step *Von Kries* chromatic
        adaptation transform with support for degree of adaptation produces
        values closer to the supplemental document compared to the
        *Zhai and Luo (2018)* two-steps chromatic adaptation transform but then
        the *ZCAM* colour appearance model does not round-trip properly.
    -   *Step 4* of the inverse model uses a rounded exponent of 1.3514
        preventing the model to round-trip properly. Given that this
        implementation takes some liberties with respect to the chromatic
        adaptation transform to use, it was deemed appropriate to use an
        exponent value that enables the *ZCAM* colour appearance model to
        round-trip.
    -   The underlying *SMPTE ST 2084:2014* transfer function is an absolute
        transfer function, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +-------------------------------+-----------------------+---------------+
    | **Domain**                    | **Scale - Reference** | **Scale - 1** |
    +===============================+=======================+===============+
    | ``CAM_Specification_ZCAM.J``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.C``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.h``  | [0, 360]              | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.s``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.Q``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.M``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.H``  | [0, 400]              | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.HC`` | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.V``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.K``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+
    | ``CAM_Specification_ZCAM.H``  | [UN]                  | [0, 1]        |
    +-------------------------------+-----------------------+---------------+

    +-----------+-----------------------+---------------+
    | **Range** | **Scale - Reference** | **Scale - 1** |
    +===========+=======================+===============+
    | ``XYZ``   | [UN]                  | [UN]          |
    +-----------+-----------------------+---------------+

    References
    ----------
    :cite:`Safdar2018`, :cite:`Safdar2021`, :cite:`Zhai2018`

    Examples
    --------
    >>> specification = CAM_Specification_ZCAM(J=92.250443780723629,
    ...                                        C=3.0216926733329013,
    ...                                        h=196.32457375575581)
    >>> XYZ_w = np.array([256, 264, 202])
    >>> L_A = 264
    >>> Y_b = 100
    >>> surround = VIEWING_CONDITIONS_ZCAM['Average']
    >>> ZCAM_to_XYZ(specification, XYZ_w, L_A, Y_b, surround)
    ... # doctest: +ELLIPSIS
    array([ 185.,  206.,  163.])
    """

    J_z, C_z, h_z, _S_z, _Q_z, M_z, _H, _H_Z, _V_z, _K_z, _W_z = astuple(
        specification
    )

    J_z = to_domain_1(J_z)
    C_z = to_domain_1(C_z)
    h_z = to_domain_degrees(h_z)
    M_z = to_domain_1(M_z)

    XYZ_w = to_domain_1(XYZ_w)
    _X_w, Y_w, _Z_w = tsplit(XYZ_w)
    L_A = as_float_array(L_A)
    Y_b = as_float_array(Y_b)

    F_s, F, c, N_c = surround

    # Step 0 (Forward) - Chromatic adaptation from reference illuminant to
    # "CIE Standard Illuminant D65" illuminant using "CAT02".
    # Computing degree of adaptation :math:`D`.
    D = (
        degree_of_adaptation(surround.F, L_A)
        if not discount_illuminant
        else ones(L_A.shape)
    )

    # Step 1 (Forward) - Computing factors related with viewing conditions and
    # independent of the test stimulus.
    # Background factor :math:`F_b`
    F_b = np.sqrt(Y_b / Y_w)
    # Luminance level adaptation factor :math:`F_L`
    F_L = 0.171 * spow(L_A, 1 / 3) * (1 - np.exp(-48 / 9 * L_A))

    # Step 2 (Forward) - Computing achromatic response (:math:`I_{z,w}`),
    # redness-greenness (:math:`a_{z,w}`), and yellowness-blueness
    # (:math:`b_{z,w}`).
    with domain_range_scale("ignore"):
        I_z_w, _A_z_w, B_z_w = tsplit(
            XYZ_to_Izazbz(XYZ_w, method="Safdar 2021")
        )

    # Step 1 (Inverse) - Computing achromatic response (:math:`I_z`).
    Q_z_p = (1.6 * F_s) / F_b**0.12
    Q_z_m = F_s**2.2 * F_b**0.5 * spow(F_L, 0.2)
    Q_z_w = 2700 * spow(I_z_w, Q_z_p) * Q_z_m

    I_z_p = (F_b**0.12) / (1.6 * F_s)
    I_z_d = 2700 * 100 * Q_z_m

    I_z = spow((J_z * Q_z_w) / I_z_d, I_z_p)

    # Step 2 (Inverse) - Computing chroma :math:`C_z`.
    if has_only_nan(M_z) and not has_only_nan(C_z):
        M_z = (C_z * Q_z_w) / 100
    elif has_only_nan(M_z):
        raise ValueError(
            'Either "C" or "M" correlate must be defined in '
            'the "CAM_Specification_ZCAM" argument!'
        )

    # Step 3 (Inverse) - Computing hue angle :math:`h_z`
    # :math:`h_z` is currently required as an input.

    # Computing eccentricity factor :math:`e_z`.
    e_z = 1.015 + np.cos(np.radians(89.038 + h_z % 360))
    h_z_r = np.radians(h_z)

    # Step 4 (Inverse) - Computing redness-greenness (:math:`a_z`), and
    # yellowness-blueness (:math:`b_z`).
    # C_z_p_e = 1.3514
    C_z_p_e = 50 / 37
    C_z_p = spow(
        (M_z * spow(I_z_w, 0.78) * F_b**0.1)
        / (100 * e_z**0.068 * spow(F_L, 0.2)),
        C_z_p_e,
    )
    a_z = C_z_p * np.cos(h_z_r)
    b_z = C_z_p * np.sin(h_z_r)

    # Step 5 (Inverse) - Computing tristimulus values :math:`XYZ_{D65}`.
    with domain_range_scale("ignore"):
        XYZ_D65 = Izazbz_to_XYZ(tstack([I_z, a_z, b_z]), method="Safdar 2021")

    XYZ = chromatic_adaptation_Zhai2018(
        XYZ_D65, TVS_D65, XYZ_w, D, D, transform="CAT02"
    )

    return from_range_1(XYZ)


def hue_quadrature(h: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the hue quadrature from given hue :math:`h` angle in degrees.

    Parameters
    ----------
    h
        Hue :math:`h` angle in degrees.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Hue quadrature.

    Examples
    --------
    >>> hue_quadrature(196.3185839)  # doctest: +ELLIPSIS
    237.6052911...
    """

    h = as_float_array(h)

    h_i = HUE_DATA_FOR_HUE_QUADRATURE["h_i"]
    e_i = HUE_DATA_FOR_HUE_QUADRATURE["e_i"]
    H_i = HUE_DATA_FOR_HUE_QUADRATURE["H_i"]

    # :math:`h_p` = :math:`h_z` + 360 if :math:`h_z` < :math:`h_1, i.e. h_i[0]
    h[h <= h_i[0]] += 360
    # *np.searchsorted* returns an erroneous index if a *nan* is used as input.
    h[np.asarray(np.isnan(h))] = 0
    i = as_int_array(np.searchsorted(h_i, h, side="left") - 1)

    h_ii = h_i[i]
    e_ii = e_i[i]
    H_ii = H_i[i]
    h_ii1 = h_i[i + 1]
    e_ii1 = e_i[i + 1]

    H = H_ii + (
        (100 * (h - h_ii) / e_ii) / ((h - h_ii) / e_ii + (h_ii1 - h) / e_ii1)
    )

    return as_float(H)



"""Defines the unit tests for the :mod:`colour.io.luts.iridas_cube` module."""

from __future__ import annotations

import numpy as np
import os
import unittest
import shutil
import tempfile

from colour.io import (
    LUT1D,
    LUTSequence,
    read_LUT_IridasCube,
    write_LUT_IridasCube,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "LUTS_DIRECTORY",
    "TestReadLUTIridasCube",
    "TestWriteLUTIridasCube",
]

LUTS_DIRECTORY: str = os.path.join(
    os.path.dirname(__file__), "resources", "iridas_cube"
)


class TestReadLUTIridasCube(unittest.TestCase):
    """
    Define :func:`colour.io.luts.iridas_cube.read_LUT_IridasCube` definition
    unit tests methods.
    """

    def test_read_LUT_IridasCube(self):
        """
        Test :func:`colour.io.luts.iridas_cube.read_LUT_IridasCube`
        definition.
        """

        LUT_1 = read_LUT_IridasCube(
            os.path.join(LUTS_DIRECTORY, "ACES_Proxy_10_to_ACES.cube")
        )

        np.testing.assert_almost_equal(
            LUT_1.table,
            np.array(
                [
                    [4.88300000e-04, 4.88300000e-04, 4.88300000e-04],
                    [7.71400000e-04, 7.71400000e-04, 7.71400000e-04],
                    [1.21900000e-03, 1.21900000e-03, 1.21900000e-03],
                    [1.92600000e-03, 1.92600000e-03, 1.92600000e-03],
                    [3.04400000e-03, 3.04400000e-03, 3.04400000e-03],
                    [4.80900000e-03, 4.80900000e-03, 4.80900000e-03],
                    [7.59900000e-03, 7.59900000e-03, 7.59900000e-03],
                    [1.20100000e-02, 1.20100000e-02, 1.20100000e-02],
                    [1.89700000e-02, 1.89700000e-02, 1.89700000e-02],
                    [2.99800000e-02, 2.99800000e-02, 2.99800000e-02],
                    [4.73700000e-02, 4.73700000e-02, 4.73700000e-02],
                    [7.48400000e-02, 7.48400000e-02, 7.48400000e-02],
                    [1.18300000e-01, 1.18300000e-01, 1.18300000e-01],
                    [1.86900000e-01, 1.86900000e-01, 1.86900000e-01],
                    [2.95200000e-01, 2.95200000e-01, 2.95200000e-01],
                    [4.66500000e-01, 4.66500000e-01, 4.66500000e-01],
                    [7.37100000e-01, 7.37100000e-01, 7.37100000e-01],
                    [1.16500000e00, 1.16500000e00, 1.16500000e00],
                    [1.84000000e00, 1.84000000e00, 1.84000000e00],
                    [2.90800000e00, 2.90800000e00, 2.90800000e00],
                    [4.59500000e00, 4.59500000e00, 4.59500000e00],
                    [7.26000000e00, 7.26000000e00, 7.26000000e00],
                    [1.14700000e01, 1.14700000e01, 1.14700000e01],
                    [1.81300000e01, 1.81300000e01, 1.81300000e01],
                    [2.86400000e01, 2.86400000e01, 2.86400000e01],
                    [4.52500000e01, 4.52500000e01, 4.52500000e01],
                    [7.15100000e01, 7.15100000e01, 7.15100000e01],
                    [1.13000000e02, 1.13000000e02, 1.13000000e02],
                    [1.78500000e02, 1.78500000e02, 1.78500000e02],
                    [2.82100000e02, 2.82100000e02, 2.82100000e02],
                    [4.45700000e02, 4.45700000e02, 4.45700000e02],
                    [7.04300000e02, 7.04300000e02, 7.04300000e02],
                ]
            ),
        )
        self.assertEqual(LUT_1.name, "ACES Proxy 10 to ACES")
        self.assertEqual(LUT_1.dimensions, 2)
        np.testing.assert_array_equal(
            LUT_1.domain, np.array([[0, 0, 0], [1, 1, 1]])
        )
        self.assertEqual(LUT_1.size, 32)
        self.assertListEqual(LUT_1.comments, [])

        LUT_2 = read_LUT_IridasCube(os.path.join(LUTS_DIRECTORY, "Demo.cube"))
        self.assertListEqual(LUT_2.comments, ["Comments can go anywhere"])
        np.testing.assert_array_equal(
            LUT_2.domain, np.array([[0, 0, 0], [1, 2, 3]])
        )

        LUT_3 = read_LUT_IridasCube(
            os.path.join(LUTS_DIRECTORY, "Three_Dimensional_Table.cube")
        )
        self.assertEqual(LUT_3.dimensions, 3)
        self.assertEqual(LUT_3.size, 2)


class TestWriteLUTIridasCube(unittest.TestCase):
    """
    Define :func:`colour.io.luts.iridas_cube.write_LUT_IridasCube` definition
    unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        self._temporary_directory = tempfile.mkdtemp()

    def tearDown(self):
        """After tests actions."""

        shutil.rmtree(self._temporary_directory)

    def test_write_LUT_IridasCube(self):
        """
        Test :func:`colour.io.luts.iridas_cube.write_LUT_IridasCube`
        definition.
        """

        LUT_1_r = read_LUT_IridasCube(
            os.path.join(LUTS_DIRECTORY, "ACES_Proxy_10_to_ACES.cube")
        )
        write_LUT_IridasCube(
            LUT_1_r,
            os.path.join(
                self._temporary_directory, "ACES_Proxy_10_to_ACES.cube"
            ),
        )
        LUT_1_t = read_LUT_IridasCube(
            os.path.join(
                self._temporary_directory, "ACES_Proxy_10_to_ACES.cube"
            )
        )
        self.assertEqual(LUT_1_r, LUT_1_t)

        write_LUT_IridasCube(
            LUTSequence(LUT_1_r),
            os.path.join(
                self._temporary_directory, "ACES_Proxy_10_to_ACES.cube"
            ),
        )
        self.assertEqual(LUT_1_r, LUT_1_t)

        LUT_2_r = read_LUT_IridasCube(
            os.path.join(LUTS_DIRECTORY, "Demo.cube")
        )
        write_LUT_IridasCube(
            LUT_2_r, os.path.join(self._temporary_directory, "Demo.cube")
        )
        LUT_2_t = read_LUT_IridasCube(
            os.path.join(self._temporary_directory, "Demo.cube")
        )
        self.assertEqual(LUT_2_r, LUT_2_t)
        self.assertListEqual(LUT_2_r.comments, LUT_2_t.comments)

        LUT_3_r = read_LUT_IridasCube(
            os.path.join(LUTS_DIRECTORY, "Three_Dimensional_Table.cube")
        )
        write_LUT_IridasCube(
            LUT_3_r,
            os.path.join(
                self._temporary_directory, "Three_Dimensional_Table.cube"
            ),
        )
        LUT_3_t = read_LUT_IridasCube(
            os.path.join(
                self._temporary_directory, "Three_Dimensional_Table.cube"
            )
        )
        self.assertEqual(LUT_3_r, LUT_3_t)

        LUT_4_r = read_LUT_IridasCube(
            os.path.join(LUTS_DIRECTORY, "ACES_Proxy_10_to_ACES.cube")
        )
        write_LUT_IridasCube(
            LUT_4_r.as_LUT(LUT1D, force_conversion=True),
            os.path.join(
                self._temporary_directory, "ACES_Proxy_10_to_ACES.cube"
            ),
        )
        LUT_4_t = read_LUT_IridasCube(
            os.path.join(
                self._temporary_directory, "ACES_Proxy_10_to_ACES.cube"
            )
        )
        self.assertEqual(LUT_4_r, LUT_4_t)


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.models.hdr_cie_lab` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.models import XYZ_to_hdr_CIELab, hdr_CIELab_to_XYZ
from colour.models.hdr_cie_lab import exponent_hdr_CIELab
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestExponent_hdr_CIELab",
    "TestXYZ_to_hdr_CIELab",
    "TestHdr_CIELab_to_XYZ",
]


class TestExponent_hdr_CIELab(unittest.TestCase):
    """
    Define :func:`colour.models.hdr_cie_lab.exponent_hdr_CIELab`
    definition unit tests methods.
    """

    def test_exponent_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.exponent_hdr_CIELab`
        definition.
        """

        self.assertAlmostEqual(
            exponent_hdr_CIELab(0.2, 100), 0.473851073746817, places=7
        )

        self.assertAlmostEqual(
            exponent_hdr_CIELab(0.4, 100), 0.656101486726362, places=7
        )

        self.assertAlmostEqual(
            exponent_hdr_CIELab(0.4, 100, method="Fairchild 2010"),
            1.326014370643925,
            places=7,
        )

        self.assertAlmostEqual(
            exponent_hdr_CIELab(0.2, 1000), 0.710776610620225, places=7
        )

    def test_n_dimensional_exponent_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.exponent_hdr_CIELab`
        definition n-dimensional arrays support.
        """

        Y_s = 0.2
        Y_abs = 100
        epsilon = exponent_hdr_CIELab(Y_s, Y_abs)

        Y_s = np.tile(Y_s, 6)
        Y_abs = np.tile(Y_abs, 6)
        epsilon = np.tile(epsilon, 6)
        np.testing.assert_almost_equal(
            exponent_hdr_CIELab(Y_s, Y_abs), epsilon, decimal=7
        )

        Y_s = np.reshape(Y_s, (2, 3))
        Y_abs = np.reshape(Y_abs, (2, 3))
        epsilon = np.reshape(epsilon, (2, 3))
        np.testing.assert_almost_equal(
            exponent_hdr_CIELab(Y_s, Y_abs), epsilon, decimal=7
        )

        Y_s = np.reshape(Y_s, (2, 3, 1))
        Y_abs = np.reshape(Y_abs, (2, 3, 1))
        epsilon = np.reshape(epsilon, (2, 3, 1))
        np.testing.assert_almost_equal(
            exponent_hdr_CIELab(Y_s, Y_abs), epsilon, decimal=7
        )

    def test_domain_range_scale_exponent_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.exponent_hdr_CIELab` definition
        domain and range scale support.
        """

        Y_s = 0.2
        Y_abs = 100
        epsilon = exponent_hdr_CIELab(Y_s, Y_abs)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    exponent_hdr_CIELab(Y_s * factor, Y_abs),
                    epsilon,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_exponent_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.exponent_hdr_CIELab`
        definition nan support.
        """

        cases = np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        exponent_hdr_CIELab(cases, cases)


class TestXYZ_to_hdr_CIELab(unittest.TestCase):
    """
    Define :func:`colour.models.hdr_cie_lab.XYZ_to_hdr_CIELab` definition unit
    tests methods.
    """

    def test_XYZ_to_hdr_CIELab(self):
        """Test :func:`colour.models.hdr_cie_lab.XYZ_to_hdr_CIELab` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(np.array([0.20654008, 0.12197225, 0.05136952])),
            np.array([51.87002062, 60.47633850, 32.14551912]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                np.array([0.44757, 0.40745]),
            ),
            np.array([51.87002062, 44.49667330, -6.69619196]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(
                np.array([0.20654008, 0.12197225, 0.05136952]),
                np.array([0.44757, 0.40745]),
                method="Fairchild 2010",
            ),
            np.array([31.99621114, 95.08564341, -14.14047055]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(
                np.array([0.20654008, 0.12197225, 0.05136952]), Y_s=0.5
            ),
            np.array([23.10388654, 59.31425004, 23.69960142]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(
                np.array([0.20654008, 0.12197225, 0.05136952]), Y_abs=1000
            ),
            np.array([29.77261805, 62.58315675, 27.31232673]),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.XYZ_to_hdr_CIELab` definition
        n-dimensional support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        illuminant = np.array([0.31270, 0.32900])
        Y_s = 0.2
        Y_abs = 100
        Lab_hdr = XYZ_to_hdr_CIELab(XYZ, illuminant, Y_s, Y_abs)

        XYZ = np.tile(XYZ, (6, 1))
        Lab_hdr = np.tile(Lab_hdr, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(XYZ, illuminant, Y_s, Y_abs), Lab_hdr, decimal=7
        )

        illuminant = np.tile(illuminant, (6, 1))
        Y_s = np.tile(Y_s, 6)
        Y_abs = np.tile(Y_abs, 6)
        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(XYZ, illuminant, Y_s, Y_abs), Lab_hdr, decimal=7
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        illuminant = np.reshape(illuminant, (2, 3, 2))
        Y_s = np.reshape(Y_s, (2, 3))
        Y_abs = np.reshape(Y_abs, (2, 3))
        Lab_hdr = np.reshape(Lab_hdr, (2, 3, 3))
        np.testing.assert_almost_equal(
            XYZ_to_hdr_CIELab(XYZ, illuminant, Y_s, Y_abs), Lab_hdr, decimal=7
        )

    def test_domain_range_scale_XYZ_to_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.XYZ_to_hdr_CIELab` definition
        domain and range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        illuminant = np.array([0.31270, 0.32900])
        Y_s = 0.2
        Y_abs = 100
        Lab_hdr = XYZ_to_hdr_CIELab(XYZ, illuminant, Y_s, Y_abs)

        d_r = (("reference", 1, 1), ("1", 1, 0.01), ("100", 100, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_hdr_CIELab(
                        XYZ * factor_a, illuminant, Y_s * factor_a, Y_abs
                    ),
                    Lab_hdr * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_hdr_CIELab(self):
        """
        Test :func:`colour.models.hdr_cie_lab.XYZ_to_hdr_CIELab` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            illuminant = np.array(case[0:2])
            Y_s = case[0]
            Y_abs = case[0]
            XYZ_to_hdr_CIELab(XYZ, illuminant, Y_s, Y_abs)


class TestHdr_CIELab_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.hdr_cie_lab.hdr_CIELab_to_XYZ` definition unit
    tests methods.
    """

    def test_hdr_CIELab_to_XYZ(self):
        """Test :func:`colour.models.hdr_cie_lab.hdr_CIELab_to_XYZ` definition."""

        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(
                np.array([51.87002062, 60.47633850, 32.14551912])
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(
                np.array([51.87002062, 44.49667330, -6.69619196]),
                np.array([0.44757, 0.40745]),
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(
                np.array([31.99621114, 95.08564341, -14.14047055]),
                np.array([0.44757, 0.40745]),
                method="Fairchild 2010",
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(
                np.array([23.10388654, 59.31425004, 23.69960142]), Y_s=0.5
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(
                np.array([29.77261805, 62.58315675, 27.31232673]), Y_abs=1000
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            decimal=7,
        )

    def test_n_dimensional_hdr_CIELab_to_XYZ(self):
        """
        Test :func:`colour.models.hdr_cie_lab.hdr_CIELab_to_XYZ` definition
        n-dimensional support.
        """

        Lab_hdr = np.array([51.87002062, 60.47633850, 32.14551912])
        illuminant = np.array([0.31270, 0.32900])
        Y_s = 0.2
        Y_abs = 100
        XYZ = hdr_CIELab_to_XYZ(Lab_hdr, illuminant, Y_s, Y_abs)

        Lab_hdr = np.tile(Lab_hdr, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(Lab_hdr, illuminant, Y_s, Y_abs), XYZ, decimal=7
        )

        illuminant = np.tile(illuminant, (6, 1))
        Y_s = np.tile(Y_s, 6)
        Y_abs = np.tile(Y_abs, 6)
        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(Lab_hdr, illuminant, Y_s, Y_abs), XYZ, decimal=7
        )

        Lab_hdr = np.reshape(Lab_hdr, (2, 3, 3))
        illuminant = np.reshape(illuminant, (2, 3, 2))
        Y_s = np.reshape(Y_s, (2, 3))
        Y_abs = np.reshape(Y_abs, (2, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_almost_equal(
            hdr_CIELab_to_XYZ(Lab_hdr, illuminant, Y_s, Y_abs), XYZ, decimal=7
        )

    def test_domain_range_scale_hdr_CIELab_to_XYZ(self):
        """
        Test :func:`colour.models.hdr_cie_lab.hdr_CIELab_to_XYZ` definition
        domain and range scale support.
        """

        Lab_hdr = np.array([26.46461067, -24.61332600, -4.84796811])
        illuminant = np.array([0.31270, 0.32900])
        Y_s = 0.2
        Y_abs = 100
        XYZ = hdr_CIELab_to_XYZ(Lab_hdr, illuminant, Y_s, Y_abs)

        d_r = (("reference", 1, 1, 1), ("1", 0.01, 1, 1), ("100", 1, 100, 100))
        for scale, factor_a, factor_b, factor_c in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    hdr_CIELab_to_XYZ(
                        Lab_hdr * factor_a, illuminant, Y_s * factor_b, Y_abs
                    ),
                    XYZ * factor_c,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_hdr_CIELab_to_XYZ(self):
        """
        Test :func:`colour.models.hdr_cie_lab.hdr_CIELab_to_XYZ` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Lab_hdr = np.array(case)
            illuminant = np.array(case[0:2])
            Y_s = case[0]
            Y_abs = case[0]
            hdr_CIELab_to_XYZ(Lab_hdr, illuminant, Y_s, Y_abs)


if __name__ == "__main__":
    unittest.main()



"""
Defines the unit tests for the
:mod:`colour.models.rgb.transfer_functions.sony_slog` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    log_encoding_SLog,
    log_decoding_SLog,
    log_encoding_SLog2,
    log_decoding_SLog2,
    log_encoding_SLog3,
    log_decoding_SLog3,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLogEncoding_SLog",
    "TestLogDecoding_SLog",
    "TestLogEncoding_SLog2",
    "TestLogDecoding_SLog2",
    "TestLogEncoding_SLog3",
    "TestLogDecoding_SLog3",
]


class TestLogEncoding_SLog(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog` definition unit tests methods.
    """

    def test_log_encoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog` definition.
        """

        self.assertAlmostEqual(
            log_encoding_SLog(0.0), 0.088251291513446, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog(0.18), 0.384970815928670, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog(0.18, 12), 0.384688786026891, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog(0.18, 10, False), 0.376512722254600, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog(0.18, 10, False, False),
            0.359987846422154,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_SLog(1.0), 0.638551684622532, places=7
        )

    def test_n_dimensional_log_encoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog` definition n-dimensional arrays support.
        """

        x = 0.18
        y = log_encoding_SLog(x)

        x = np.tile(x, 6)
        y = np.tile(y, 6)
        np.testing.assert_almost_equal(log_encoding_SLog(x), y, decimal=7)

        x = np.reshape(x, (2, 3))
        y = np.reshape(y, (2, 3))
        np.testing.assert_almost_equal(log_encoding_SLog(x), y, decimal=7)

        x = np.reshape(x, (2, 3, 1))
        y = np.reshape(y, (2, 3, 1))
        np.testing.assert_almost_equal(log_encoding_SLog(x), y, decimal=7)

    def test_domain_range_scale_log_encoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog` definition domain and range scale support.
        """

        x = 0.18
        y = log_encoding_SLog(x)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_SLog(x * factor), y * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog` definition nan support.
        """

        log_encoding_SLog(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLogDecoding_SLog(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog` definition unit tests methods.
    """

    def test_log_decoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog` definition.
        """

        self.assertAlmostEqual(
            log_decoding_SLog(0.088251291513446), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog(0.384970815928670), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog(0.384688786026891, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog(0.376512722254600, 10, False), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog(0.359987846422154, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_SLog(0.638551684622532), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog` definition n-dimensional arrays support.
        """

        y = 0.384970815928670
        x = log_decoding_SLog(y)

        y = np.tile(y, 6)
        x = np.tile(x, 6)
        np.testing.assert_almost_equal(log_decoding_SLog(y), x, decimal=7)

        y = np.reshape(y, (2, 3))
        x = np.reshape(x, (2, 3))
        np.testing.assert_almost_equal(log_decoding_SLog(y), x, decimal=7)

        y = np.reshape(y, (2, 3, 1))
        x = np.reshape(x, (2, 3, 1))
        np.testing.assert_almost_equal(log_decoding_SLog(y), x, decimal=7)

    def test_domain_range_scale_log_decoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog` definition domain and range scale support.
        """

        y = 0.384970815928670
        x = log_decoding_SLog(y)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_SLog(y * factor), x * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_SLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog` definition nan support.
        """

        log_decoding_SLog(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLogEncoding_SLog2(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog2` definition unit tests methods.
    """

    def test_log_encoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog2` definition.
        """

        self.assertAlmostEqual(
            log_encoding_SLog2(0.0), 0.088251291513446, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog2(0.18), 0.339532524633774, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog2(0.18, 12), 0.339283782857486, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog2(0.18, 10, False), 0.323449512215013, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog2(0.18, 10, False, False),
            0.307980741258647,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_SLog2(1.0), 0.585091059564112, places=7
        )

    def test_n_dimensional_log_encoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog2` definition n-dimensional arrays support.
        """

        x = 0.18
        y = log_encoding_SLog2(x)

        x = np.tile(x, 6)
        y = np.tile(y, 6)
        np.testing.assert_almost_equal(log_encoding_SLog2(x), y, decimal=7)

        x = np.reshape(x, (2, 3))
        y = np.reshape(y, (2, 3))
        np.testing.assert_almost_equal(log_encoding_SLog2(x), y, decimal=7)

        x = np.reshape(x, (2, 3, 1))
        y = np.reshape(y, (2, 3, 1))
        np.testing.assert_almost_equal(log_encoding_SLog2(x), y, decimal=7)

    def test_domain_range_scale_log_encoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog2` definition domain and range scale support.
        """

        x = 0.18
        y = log_encoding_SLog2(x)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_SLog2(x * factor), y * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog2` definition nan support.
        """

        log_encoding_SLog2(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLogDecoding_SLog2(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog2` definition unit tests methods.
    """

    def test_log_decoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog2` definition.
        """

        self.assertAlmostEqual(
            log_decoding_SLog2(0.088251291513446), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog2(0.339532524633774), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog2(0.339283782857486, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog2(0.323449512215013, 10, False), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog2(0.307980741258647, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_SLog2(0.585091059564112), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog2` definition n-dimensional arrays support.
        """

        y = 0.339532524633774
        x = log_decoding_SLog2(y)

        y = np.tile(y, 6)
        x = np.tile(x, 6)
        np.testing.assert_almost_equal(log_decoding_SLog2(y), x, decimal=7)

        y = np.reshape(y, (2, 3))
        x = np.reshape(x, (2, 3))
        np.testing.assert_almost_equal(log_decoding_SLog2(y), x, decimal=7)

        y = np.reshape(y, (2, 3, 1))
        x = np.reshape(x, (2, 3, 1))
        np.testing.assert_almost_equal(log_decoding_SLog2(y), x, decimal=7)

    def test_domain_range_scale_log_decoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog2` definition domain and range scale support.
        """

        y = 0.339532524633774
        x = log_decoding_SLog2(y)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_SLog2(y * factor), x * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_SLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog2` definition nan support.
        """

        log_decoding_SLog2(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLogEncoding_SLog3(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog3` definition unit tests methods.
    """

    def test_log_encoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog3` definition.
        """

        self.assertAlmostEqual(
            log_encoding_SLog3(0.0), 0.092864125122190, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog3(0.18), 0.41055718475073, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog3(0.18, 12), 0.410557184750733, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog3(0.18, 10, False), 0.406392694063927, places=7
        )

        self.assertAlmostEqual(
            log_encoding_SLog3(0.18, 10, False, False),
            0.393489294768447,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_SLog3(1.0), 0.596027343690123, places=7
        )

    def test_n_dimensional_log_encoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog3` definition n-dimensional arrays support.
        """

        x = 0.18
        y = log_encoding_SLog3(x)

        x = np.tile(x, 6)
        y = np.tile(y, 6)
        np.testing.assert_almost_equal(log_encoding_SLog3(x), y, decimal=7)

        x = np.reshape(x, (2, 3))
        y = np.reshape(y, (2, 3))
        np.testing.assert_almost_equal(log_encoding_SLog3(x), y, decimal=7)

        x = np.reshape(x, (2, 3, 1))
        y = np.reshape(y, (2, 3, 1))
        np.testing.assert_almost_equal(log_encoding_SLog3(x), y, decimal=7)

    def test_domain_range_scale_log_encoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog3` definition domain and range scale support.
        """

        x = 0.18
        y = log_encoding_SLog3(x)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_SLog3(x * factor), y * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_encoding_SLog3` definition nan support.
        """

        log_encoding_SLog3(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLogDecoding_SLog3(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog3` definition unit tests methods.
    """

    def test_log_decoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog3` definition.
        """

        self.assertAlmostEqual(
            log_decoding_SLog3(0.092864125122190), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog3(0.41055718475073), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog3(0.410557184750733, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog3(0.406392694063927, 10, False), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_SLog3(0.393489294768447, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_SLog3(0.596027343690123), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog3` definition n-dimensional arrays support.
        """

        y = 0.41055718475073
        x = log_decoding_SLog3(y)

        y = np.tile(y, 6)
        x = np.tile(x, 6)
        np.testing.assert_almost_equal(log_decoding_SLog3(y), x, decimal=7)

        y = np.reshape(y, (2, 3))
        x = np.reshape(x, (2, 3))
        np.testing.assert_almost_equal(log_decoding_SLog3(y), x, decimal=7)

        y = np.reshape(y, (2, 3, 1))
        x = np.reshape(x, (2, 3, 1))
        np.testing.assert_almost_equal(log_decoding_SLog3(y), x, decimal=7)

    def test_domain_range_scale_log_decoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog3` definition domain and range scale support.
        """

        y = 0.41055718475073
        x = log_decoding_SLog3(y)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_SLog3(y * factor), x * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_SLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.sony_slog.\
log_decoding_SLog3` definition nan support.
        """

        log_decoding_SLog3(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.colorimetry.correction` module."""

import numpy as np
import unittest

from colour.colorimetry import (
    SpectralDistribution,
    bandpass_correction_Stearns1988,
)
from colour.hints import Tuple

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "DATA_NON_BANDPASS_CORRECTED",
    "DATA_BANDPASS_CORRECTED",
    "TestBandpassCorrectionStearns1988",
]

DATA_NON_BANDPASS_CORRECTED: Tuple = (
    9.3700,
    12.3200,
    12.4600,
    9.5100,
    5.9200,
    4.3300,
    4.2900,
    3.8800,
    4.5100,
    10.9200,
    27.5000,
    49.6700,
    69.5900,
    81.7300,
    88.1900,
    86.0500,
)

DATA_BANDPASS_CORRECTED: Tuple = (
    9.12515000,
    12.57355255,
    12.69542514,
    9.54357971,
    5.75121288,
    4.21535933,
    4.33022518,
    3.79034131,
    4.03770167,
    10.11509076,
    27.10283747,
    49.88971449,
    70.21750370,
    82.14935719,
    88.88373581,
    85.87238000,
)


class TestBandpassCorrectionStearns1988(unittest.TestCase):
    """
    Define :func:`colour.colorimetry.correction.\
bandpass_correction_Stearns1988` definition unit tests methods.
    """

    def test_bandpass_correction_Stearns1988(self):
        """
        Test :func:`colour.colorimetry.correction.\
bandpass_correction_Stearns1988` definition.
        """

        sd = SpectralDistribution(
            dict(
                zip(
                    range(len(DATA_NON_BANDPASS_CORRECTED)),
                    DATA_NON_BANDPASS_CORRECTED,
                )
            )
        )

        np.testing.assert_almost_equal(
            bandpass_correction_Stearns1988(sd).values, DATA_BANDPASS_CORRECTED
        )


if __name__ == "__main__":
    unittest.main()



"""
:math:`IC_AC_B` Colourspace
===========================

Defines the :math:`IC_AC_B` colourspace transformations:

-   :func:`colour.XYZ_to_ICaCb`
-   :func:`colour.ICaCb_to_XYZ`

References
----------
-   :cite:`Frohlich2017` : Fröhlich, J. (2017). Encoding high dynamic range
    and wide color gamut imagery. doi:10.18419/OPUS-9664
"""

from __future__ import annotations

import numpy as np

from colour.algebra import vector_dot
from colour.hints import ArrayLike, NDArray
from colour.models.rgb.transfer_functions import (
    eotf_ST2084,
    eotf_inverse_ST2084,
)
from colour.utilities import (
    domain_range_scale,
    from_range_1,
    to_domain_1,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "MATRIX_ICACB_XYZ_TO_LMS",
    "MATRIX_ICACB_LMS_TO_XYZ",
    "MATRIX_ICACB_XYZ_TO_LMS_2",
    "MATRIX_ICACB_LMS_TO_XYZ_2",
    "XYZ_to_ICaCb",
    "ICaCb_to_XYZ",
]

MATRIX_ICACB_XYZ_TO_LMS: NDArray = np.array(
    [
        [0.37613, 0.70431, -0.05675],
        [-0.21649, 1.14744, 0.05356],
        [0.02567, 0.16713, 0.74235],
    ]
)
"""*CIE XYZ* tristimulus values to normalised cone responses matrix."""

MATRIX_ICACB_LMS_TO_XYZ: NDArray = np.linalg.inv(MATRIX_ICACB_XYZ_TO_LMS)
"""Normalised cone responses to *CIE XYZ* tristimulus values matrix."""

MATRIX_ICACB_XYZ_TO_LMS_2: NDArray = np.array(
    [
        [0.4949, 0.5037, 0.0015],
        [4.2854, -4.5462, 0.2609],
        [0.3605, 1.1499, -1.5105],
    ]
)
"""Normalised non-linear cone responses to :math:`IC_AC_B` colourspace matrix."""

MATRIX_ICACB_LMS_TO_XYZ_2: NDArray = np.linalg.inv(MATRIX_ICACB_XYZ_TO_LMS_2)
""":math:`IC_AC_B` to normalised non-linear cone responses colourspace matrix."""


def XYZ_to_ICaCb(XYZ: ArrayLike) -> NDArray:
    """
    Convert from *CIE XYZ* tristimulus values to :math:`IC_AC_B` colourspace.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        :math:`IC_AC_B` colourspace array.

    Notes
    -----
    +------------+-----------------------+-----------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``XYZ``    | [0, 1]                | [0, 1]          |
    +------------+-----------------------+-----------------+

    +------------+-----------------------+-----------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``ICaCb``  | ``I`` : [0, 1]        | ``I`` : [0, 1]  |
    |            |                       |                 |
    |            | ``Ca`` : [-1, 1]      | ``Ca``: [-1, 1] |
    |            |                       |                 |
    |            | ``Cb`` : [-1, 1]      | ``Cb``: [-1, 1] |
    +------------+-----------------------+-----------------+

    -   Input *CIE XYZ* tristimulus values must be adapted to
        *CIE Standard Illuminant D Series* *D65*.

    References
    ----------
    :cite:`Frohlich2017`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_to_ICaCb(XYZ)
    array([ 0.06875297,  0.05753352,  0.02081548])
    """

    XYZ = to_domain_1(XYZ)
    LMS = vector_dot(MATRIX_ICACB_XYZ_TO_LMS, XYZ)

    with domain_range_scale("ignore"):
        LMS_prime = eotf_inverse_ST2084(LMS)

    return from_range_1(vector_dot(MATRIX_ICACB_XYZ_TO_LMS_2, LMS_prime))


def ICaCb_to_XYZ(ICaCb: ArrayLike) -> NDArray:
    """
    Convert from :math:`IC_AC_B` tristimulus values to *CIE XYZ* colourspace.

    Parameters
    ----------
    ICaCb
        :math:`IC_AC_B` tristimulus values.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ* colourspace array.

    Notes
    -----
    +------------+-----------------------+-----------------+
    | **Domain** | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``ICaCb``  | ``I`` : [0, 1]        | ``I`` : [0, 1]  |
    |            |                       |                 |
    |            | ``Ca`` : [-1, 1]      | ``Ca``: [-1, 1] |
    |            |                       |                 |
    |            | ``Cb`` : [-1, 1]      | ``Cb``: [-1, 1] |
    +------------+-----------------------+-----------------+

    +------------+-----------------------+-----------------+
    | **Range**  | **Scale - Reference** | **Scale - 1**   |
    +============+=======================+=================+
    | ``XYZ``    | [0, 1]                | [0, 1]          |
    +------------+-----------------------+-----------------+

    References
    ----------
    :cite:`Frohlich2017`

    Examples
    --------
    >>> XYZ = np.array([0.06875297, 0.05753352, 0.02081548])
    >>> ICaCb_to_XYZ(XYZ)
    array([ 0.20654008,  0.12197225,  0.05136951])
    """

    ICaCb = to_domain_1(ICaCb)
    LMS_prime = vector_dot(MATRIX_ICACB_LMS_TO_XYZ_2, ICaCb)

    with domain_range_scale("ignore"):
        LMS = eotf_ST2084(LMS_prime)

    return from_range_1(vector_dot(MATRIX_ICACB_LMS_TO_XYZ, LMS))



"""
Helmholtz—Kohlrausch Effect
===========================

Defines the following methods for estimating Helmholtz-Kohlrausch effect (HKE):

-   :attr:`colour.HKE_NAYATANI1997_METHODS`: Nayatani HKE computation methods,
    choice between variable achromatic colour ('VAC') and variable chromatic
    colour ('VCC').
-   :func:`colour.HelmholtzKohlrausch_effect_object_Nayatani1997`:
    *Nayatani (1997)* HKE estimation for object colours.
-   :func:`colour.HelmholtzKohlrausch_effect_luminous_Nayatani1997`:
    *Nayatani (1997)* HKE estimation for luminous colours.
-   :func:`colour.appearance.coefficient_q_Nayatani1997`:
    Calculates :math:`WI` coefficient for *Nayatani 1997* HKE estimation.
-   :func:`colour.appearance.coefficient_K_Br_Nayatani1997`:
    Calculates :math:`K_{Br}` coefficient for *Nayatani 1997* HKE estimation.

References
----------
-   :cite:`Nayatani1997` : Nayatani, Y. (1997). Simple estimation methods for
    the Helmholtz—Kohlrausch effect. Color Research & Application, 22(6),
    385-401. doi:10.1002/(SICI)1520-6378(199712)22:6<385::AID-COL6>3.0.CO;2-R
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow
from colour.hints import (
    ArrayLike,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float_array,
    tsplit,
    validate_method,
)

__author__ = "Ilia Sibiryakov"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "HKE_NAYATANI1997_METHODS",
    "HelmholtzKohlrausch_effect_object_Nayatani1997",
    "HelmholtzKohlrausch_effect_luminous_Nayatani1997",
    "coefficient_q_Nayatani1997",
    "coefficient_K_Br_Nayatani1997",
]

HKE_NAYATANI1997_METHODS = CaseInsensitiveMapping(
    {
        "VAC": -0.1340,
        "VCC": -0.8660,
    }
)
HKE_NAYATANI1997_METHODS.__doc__ = """
*Nayatani (1997)* *HKE* computation methods, choice between variable achromatic
colour ('VAC') and variable chromatic colour ('VCC')

References
----------
:cite:`Nayatani1997`
"""


def HelmholtzKohlrausch_effect_object_Nayatani1997(
    uv: ArrayLike,
    uv_c: ArrayLike,
    L_a: FloatingOrArrayLike,
    method: Union[Literal["VAC", "VCC"], str] = "VCC",
) -> FloatingOrNDArray:
    """
    Return the *HKE* value for object colours using *Nayatani (1997)* method.

    Parameters
    ----------
    uv
        *CIE uv* chromaticity coordinates of samples.
    uv_c
        *CIE uv* chromaticity coordinates of reference white.
    L_a
        Adapting luminance in :math:`cd/m^2`.
    method
        Which estimation method to use, *VCC* or *VAC*.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance factor (:math:`\\Gamma`) value(s) computed with Nayatani
        object colour estimation method.

    References
    ----------
    :cite:`Nayatani1997`

    Examples
    --------
    >>> import colour
    >>> white = colour.xy_to_Luv_uv(colour.temperature.CCT_to_xy_CIE_D(6504))
    >>> colours = colour.XYZ_to_xy(
    ...     [colour.wavelength_to_XYZ(430 + i * 50) for i in range(5)])
    >>> L_adapting = 65
    >>> HelmholtzKohlrausch_effect_object_Nayatani1997(  # doctest: +ELLIPSIS
    ...     colour.xy_to_Luv_uv(colours), white, L_adapting)
    array([ 2.2468383...,  1.4619799...,  1.1801658...,  0.9031318...,  \
1.7999376...])
    """

    u, v = tsplit(uv)
    u_c, v_c = tsplit(uv_c)

    method = validate_method(method, HKE_NAYATANI1997_METHODS)

    K_Br = coefficient_K_Br_Nayatani1997(L_a)
    q = coefficient_q_Nayatani1997(np.arctan2(v - v_c, u - u_c))
    S_uv = 13 * np.sqrt((u - u_c) ** 2 + (v - v_c) ** 2)

    return 1 + (HKE_NAYATANI1997_METHODS[method] * q + 0.0872 * K_Br) * S_uv


def HelmholtzKohlrausch_effect_luminous_Nayatani1997(
    uv: ArrayLike,
    uv_c: ArrayLike,
    L_a: FloatingOrArrayLike,
    method: Union[Literal["VAC", "VCC"], str] = "VCC",
) -> FloatingOrNDArray:
    """
    Return the *HKE* factor for luminous colours using *Nayatani (1997)* method.

    Parameters
    ----------
    uv
        *CIE uv* chromaticity coordinates of samples.
    uv_c
        *CIE uv* chromaticity coordinates of reference white.
    L_a
        Adapting luminance in :math:`cd/m^2`.
    method
        Which estimation method to use, *VCC* or *VAC*.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Luminance factor (:math:`\\Gamma`) value(s) computed with Nayatani
        luminous colour estimation method.

    References
    ----------
    :cite:`Nayatani1997`

    Examples
    --------
    >>> import colour
    >>> white = colour.xy_to_Luv_uv(colour.temperature.CCT_to_xy_CIE_D(6504))
    >>> colours = colour.XYZ_to_xy(
    ...     [colour.wavelength_to_XYZ(430 + i * 50) for i in range(5)])
    >>> L_adapting = 65
    >>> HelmholtzKohlrausch_effect_luminous_Nayatani1997(  # doctest: +ELLIPSIS
    ...     colour.xy_to_Luv_uv(colours), white, L_adapting)
    array([ 7.4460471...,  2.4767159...,  1.4723422...,  0.7938695...,  \
4.1828629...])
    """

    return (
        0.4462
        * (
            HelmholtzKohlrausch_effect_object_Nayatani1997(
                uv, uv_c, L_a, method
            )
            + 0.3086
        )
        ** 3
    )


def coefficient_q_Nayatani1997(
    theta: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the :math:`q(\\theta)` coefficient for *Nayatani (1997)* *HKE*
    computations.

    The hue angle :math:`\\theta` can be computed as follows:

    :math:`tan^{-1}\\cfrac{v' - v'_c}{u' - u'_c}`

    where :math:`u'` and :math:`v'` are the CIE 1976 chromaticity coordinates
    of the test chromatic light and :math:`u'_c` and :math:`v'_c` are the CIE
    1976 chromaticity coordinates of the reference white light.

    Parameters
    ----------
    theta
        Hue angle (:math:`\\theta`) in radians.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`q` coefficient for *Nayatani (1997)* *HKE* methods.

    References
    ----------
    :cite:`Nayatani1997`

    Examples
    --------
    This recreates *FIG. A-1*.

    >>> import matplotlib.pyplot as plt
    >>> angles = [(np.pi * 2 / 100 * i) for i in range(100)]
    >>> q_values = coefficient_q_Nayatani1997(angles)
    >>> plt.plot(np.array(angles), q_values / (np.pi * 2) * 180)
    ... # doctest: +ELLIPSIS
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.show()  # doctest: +SKIP
    """

    theta = as_float_array(theta)

    theta_2, theta_3, theta_4 = 2 * theta, 3 * theta, 4 * theta

    return (
        -0.01585
        - 0.03017 * np.cos(theta)
        - 0.04556 * np.cos(theta_2)
        - 0.02667 * np.cos(theta_3)
        - 0.00295 * np.cos(theta_4)
        + 0.14592 * np.sin(theta)
        + 0.05084 * np.sin(theta_2)
        - 0.01900 * np.sin(theta_3)
        - 0.00764 * np.sin(theta_4)
    )


def coefficient_K_Br_Nayatani1997(
    L_a: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the :math:`K_{Br}` coefficient for *Nayatani (1997)* *HKE*
    computations.

    Parameters
    ----------
    L_a
        Adapting luminance in :math:`cd/m^2`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        :math:`K_{Br}` coefficient for *Nayatani (1997)* *HKE* methods.

    Notes
    -----
    -   The :math:`K_{Br}` coefficient is normalised to unity around
        :math:`63.66cd/m^2`.

    References
    ----------
    :cite:`Nayatani1997`

    Examples
    --------
    >>> L_a_values = [10 + i * 20 for i in range(5)]
    >>> coefficient_K_Br_Nayatani1997(L_a_values)  # doctest: +ELLIPSIS
    array([ 0.7134481...,  0.8781172...,  0.9606248...,  1.0156689...,  \
1.0567008...])
    >>> coefficient_K_Br_Nayatani1997(63.66)  # doctest: +ELLIPSIS
    1.0001284...
    """

    L_a_4495 = spow(L_a, 0.4495)

    return 0.2717 * (6.469 + 6.362 * L_a_4495) / (6.469 + L_a_4495)



"""Defines the unit tests for the :mod:`colour.utilities.verbose` module."""

import os
import sys
import unittest

from colour.utilities import (
    show_warning,
    suppress_warnings,
    describe_environment,
)
from colour.utilities import warning

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestShowWarning",
    "TestSuppressWarnings",
    "TestDescribeEnvironment",
]


class TestShowWarning(unittest.TestCase):
    """
    Define :func:`colour.utilities.verbose.show_warning` definition unit tests
    methods.
    """

    def test_show_warning(self):
        """Test :func:`colour.utilities.verbose.show_warning` definition."""

        show_warning("This is a unit test warning!", Warning, None, None)

        with open(os.devnull) as dev_null:
            show_warning(
                "This is a unit test warning!", Warning, None, None, dev_null
            )

        stderr = sys.stderr
        try:
            sys.stderr = None
            show_warning("This is a unit test warning!", Warning, None, None)
        finally:
            sys.stderr = stderr


class TestSuppressWarnings(unittest.TestCase):
    """
    Define :func:`colour.utilities.verbose.suppress_warnings` definition unit
    tests methods.
    """

    def test_suppress_warnings(self):
        """Test :func:`colour.utilities.verbose.suppress_warnings` definition."""

        with suppress_warnings():
            warning("This is a suppressed unit test warning!")


class TestDescribeEnvironment(unittest.TestCase):
    """
    Define :func:`colour.utilities.verbose.describe_environment` definition
    unit tests methods.
    """

    def test_describe_environment(self):
        """Test :func:`colour.utilities.verbose.describe_environment` definition."""

        environment = describe_environment()
        self.assertIsInstance(environment, dict)
        self.assertListEqual(
            sorted(environment.keys()),
            ["Interpreter", "Runtime", "colour-science.org"],
        )

        environment = describe_environment(development_packages=True)
        self.assertListEqual(
            sorted(environment.keys()),
            ["Development", "Interpreter", "Runtime", "colour-science.org"],
        )

        environment = describe_environment(
            development_packages=True, extras_packages=True
        )
        self.assertListEqual(
            sorted(environment.keys()),
            [
                "Development",
                "Extras",
                "Interpreter",
                "Runtime",
                "colour-science.org",
            ],
        )


if __name__ == "__main__":
    unittest.main()



"""Showcases dominant wavelength and purity of a colour computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box("Dominant Wavelength and Purity")

xy = np.array([0.54369557, 0.32107944])
xy_n = np.array([0.31270, 0.32900])
cmfs = colour.MSDS_CMFS["CIE 1931 2 Degree Standard Observer"]
message_box(
    f'Computing the "dominant wavelength" for colour stimulus "xy" and '
    f'achromatic stimulus "xy_n" chromaticity coordinates:\n\n'
    f"\txy: {xy}\n"
    f"\txy_n: {xy_n}"
)

print(colour.dominant_wavelength(xy, xy_n, cmfs))

print("\n")

xy = np.array([0.35000, 0.25000])
message_box(
    f'Computing the "dominant wavelength" for colour stimulus "xy" and '
    f'achromatic stimulus "xy_n" chromaticity coordinates:\n\n'
    f"\txy: {xy}\n"
    f"\txy_n: {xy_n}\n\n"
    f'In this case the "complementary dominant wavelength" indicated by a '
    f"negative sign is returned because the first intersection is located on "
    f"the line of purples."
)

print(colour.dominant_wavelength(xy, xy_n, cmfs))

print("\n")

xy = np.array([0.54369557, 0.32107944])
message_box(
    f'Computing the "complementary wavelength" for colour stimulus "xy" and '
    f'achromatic stimulus "xy_n" chromaticity coordinates:\n\n'
    f"\txy: {xy}\n"
    f"\txy_n: {xy_n}\n\n"
    f'In this case the "dominant wavelength" indicated by a negative sign is '
    f"returned because the first intersection is located on the line of purples."
)

print(colour.complementary_wavelength(xy, xy_n, cmfs))

print("\n")

xy = np.array([0.35000, 0.25000])
message_box(
    f'Computing the "complementary wavelength" for colour stimulus "xy" and '
    f'achromatic stimulus "xy_n" chromaticity coordinates:\n\n'
    f"\txy: {xy}\n"
    f"\txy_n: {xy_n}"
)

print(colour.complementary_wavelength(xy, xy_n, cmfs))

print("\n")

xy = np.array([0.54369557, 0.32107944])
message_box(
    f'Computing the "excitation purity" for colour stimulus "xy" and '
    f'achromatic stimulus "xy_n" chromaticity coordinates:\n\n'
    f"\txy: {xy}\n"
    f"\txy_n: {xy_n}"
)

print(colour.excitation_purity(xy, xy_n, cmfs))

print("\n")

xy = np.array([0.54369557, 0.32107944])
message_box(
    f'Computing the "colorimetric purity" for colour stimulus "xy" and '
    f'achromatic stimulus "xy_n" chromaticity coordinates:\n\n'
    f"\txy: {xy}\n"
    f"\txy_n: {xy_n}"
)

print(colour.colorimetric_purity(xy, xy_n, cmfs))



"""
SRGB Colourspace
================

Defines the *sRGB* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_sRGB`.

References
----------
-   :cite:`InternationalElectrotechnicalCommission1999a` : International
    Electrotechnical Commission. (1999). IEC 61966-2-1:1999 - Multimedia
    systems and equipment - Colour measurement and management - Part 2-1:
    Colour management - Default RGB colour space - sRGB (p. 51).
    https://webstore.iec.ch/publication/6169
-   :cite:`InternationalTelecommunicationUnion2015i` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.709-6 - Parameter
    values for the HDTV standards for production and international programme
    exchange BT Series Broadcasting service (pp. 1-32).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.709-6-201506-I!!PDF-E.pdf
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import RGB_Colourspace, eotf_inverse_sRGB, eotf_sRGB

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_sRGB",
    "WHITEPOINT_NAME_sRGB",
    "CCS_WHITEPOINT_sRGB",
    "MATRIX_sRGB_TO_XYZ",
    "MATRIX_XYZ_TO_sRGB",
    "RGB_COLOURSPACE_sRGB",
]

PRIMARIES_sRGB: NDArray = np.array(
    [
        [0.6400, 0.3300],
        [0.3000, 0.6000],
        [0.1500, 0.0600],
    ]
)
"""*sRGB* colourspace primaries."""

WHITEPOINT_NAME_sRGB: str = "D65"
"""*sRGB* colourspace whitepoint name."""

CCS_WHITEPOINT_sRGB: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_sRGB]
"""*sRGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_sRGB_TO_XYZ: NDArray = np.array(
    [
        [0.4124, 0.3576, 0.1805],
        [0.2126, 0.7152, 0.0722],
        [0.0193, 0.1192, 0.9505],
    ]
)
"""*sRGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_sRGB: NDArray = np.array(
    [
        [3.2406, -1.5372, -0.4986],
        [-0.9689, 1.8758, 0.0415],
        [0.0557, -0.2040, 1.0570],
    ]
)
"""*CIE XYZ* tristimulus values to *sRGB* colourspace matrix."""

RGB_COLOURSPACE_sRGB: RGB_Colourspace = RGB_Colourspace(
    "sRGB",
    PRIMARIES_sRGB,
    CCS_WHITEPOINT_sRGB,
    WHITEPOINT_NAME_sRGB,
    MATRIX_sRGB_TO_XYZ,
    MATRIX_XYZ_TO_sRGB,
    eotf_inverse_sRGB,
    eotf_sRGB,
)
RGB_COLOURSPACE_sRGB.__doc__ = """
*sRGB* colourspace.

References
----------
:cite:`InternationalElectrotechnicalCommission1999a`,
:cite:`InternationalTelecommunicationUnion2015i`
"""



"""
Academy Spectral Similarity Index (SSI)
=======================================

Defines the *Academy Spectral Similarity Index* (SSI) computation objects:

-   :func:`colour.spectral_similarity_index`

References
----------
-   :cite:`TheAcademyofMotionPictureArtsandSciences2019` : The Academy of
    Motion Picture Arts and Sciences. (2019). Academy Spectral Similarity Index
    (SSI): Overview (pp. 1-7).
"""

from __future__ import annotations

import numpy as np
from scipy.ndimage.filters import convolve1d

from colour.algebra import LinearInterpolator
from colour.colorimetry import SpectralDistribution, SpectralShape, reshape_sd
from colour.hints import NDArray, Optional
from colour.utilities import zeros

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "SPECTRAL_SHAPE_SSI",
    "spectral_similarity_index",
]

SPECTRAL_SHAPE_SSI: SpectralShape = SpectralShape(375, 675, 1)
"""*Academy Spectral Similarity Index* (SSI) spectral shape."""

_SPECTRAL_SHAPE_SSI_LARGE: SpectralShape = SpectralShape(380, 670, 10)

_MATRIX_INTEGRATION: Optional[NDArray] = None


def spectral_similarity_index(
    sd_test: SpectralDistribution, sd_reference: SpectralDistribution
) -> NDArray:
    """
    Return the *Academy Spectral Similarity Index* (SSI) of given test
    spectral distribution with given reference spectral distribution.

    Parameters
    ----------
    sd_test
        Test spectral distribution.
    sd_reference
        Reference spectral distribution.

    Returns
    -------
    :class:`numpy.ndarray`
        *Academy Spectral Similarity Index* (SSI).

    References
    ----------
    :cite:`TheAcademyofMotionPictureArtsandSciences2019`

    Examples
    --------
    >>> from colour import SDS_ILLUMINANTS
    >>> sd_test = SDS_ILLUMINANTS['C']
    >>> sd_reference = SDS_ILLUMINANTS['D65']
    >>> spectral_similarity_index(sd_test, sd_reference)
    94.0
    """

    global _MATRIX_INTEGRATION

    if _MATRIX_INTEGRATION is None:
        _MATRIX_INTEGRATION = zeros(
            (
                len(_SPECTRAL_SHAPE_SSI_LARGE.range()),
                len(SPECTRAL_SHAPE_SSI.range()),
            )
        )

        weights = np.array([0.5, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0.5])

        for i in range(_MATRIX_INTEGRATION.shape[0]):
            _MATRIX_INTEGRATION[i, (10 * i) : (10 * i + 11)] = weights

    settings = {
        "interpolator": LinearInterpolator,
        "extrapolator_kwargs": {"left": 0, "right": 0},
    }

    sd_test = reshape_sd(sd_test, SPECTRAL_SHAPE_SSI, "Align", **settings)
    sd_reference = reshape_sd(
        sd_reference, SPECTRAL_SHAPE_SSI, "Align", **settings
    )

    test_i = np.dot(_MATRIX_INTEGRATION, sd_test.values)
    reference_i = np.dot(_MATRIX_INTEGRATION, sd_reference.values)

    test_i /= np.sum(test_i)
    reference_i /= np.sum(reference_i)

    d_i = test_i - reference_i
    dr_i = d_i / (reference_i + np.mean(reference_i))
    wdr_i = dr_i * [
        12 / 45,
        22 / 45,
        32 / 45,
        40 / 45,
        44 / 45,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        11 / 15,
        3 / 15,
    ]
    c_wdr_i = convolve1d(np.hstack([0, wdr_i, 0]), [0.22, 0.56, 0.22])
    m_v = np.sum(c_wdr_i**2)

    SSI = np.around(100 - 32 * np.sqrt(m_v))

    return SSI



"""
Luminance :math:`Y`
===================

Defines the *luminance* :math:`Y` computation objects.

The following methods are available:

-   :func:`colour.colorimetry.luminance_Newhall1943`: *luminance* :math:`Y`
    computation of given *Munsell* value :math:`V` using
    *Newhall, Nickerson and Judd (1943)* method.
-   :func:`colour.colorimetry.luminance_ASTMD1535`: *luminance* :math:`Y`
    computation of given *Munsell* value :math:`V` using *ASTM D1535-08e1*
    method.
-   :func:`colour.colorimetry.luminance_CIE1976`: *luminance* :math:`Y`
    computation of given *Lightness* :math:`L^*` as per *CIE 1976*
    recommendation.
-   :func:`colour.colorimetry.luminance_Fairchild2010`: *luminance* :math:`Y`
    computation of given *Lightness* :math:`L_{hdr}` using
    *Fairchild and Wyble (2010)* method.
-   :func:`colour.colorimetry.luminance_Fairchild2011`: *luminance* :math:`Y`
    computation of given *Lightness* :math:`L_{hdr}` using
    *Fairchild and Chen (2011)* method.
-   :func:`colour.colorimetry.luminance_Abebe2017`: *Luminance* :math:`Y`
    computation of given *Lightness* :math:`L` using
    *Abebe, Pouli, Larabi and Reinhard (2017)* method.
-   :attr:`colour.LUMINANCE_METHODS`: Supported *luminance* :math:`Y`
    computation methods.
-   :func:`colour.luminance`: *Luminance* :math:`Y` computation of given
    *Lightness* :math:`L^*` or given *Munsell* value :math:`V` using given
    method.

References
----------
-   :cite:`Abebe2017a` : Abebe, M. A., Pouli, T., Larabi, M.-C., & Reinhard,
    E. (2017). Perceptual Lightness Modeling for High-Dynamic-Range Imaging.
    ACM Transactions on Applied Perception, 15(1), 1-19. doi:10.1145/3086577
-   :cite:`ASTMInternational2008a` : ASTM International. (2008). ASTM
    D1535-08e1 - Standard Practice for Specifying Color by the Munsell System.
    doi:10.1520/D1535-08E01
-   :cite:`CIETC1-482004m` : CIE TC 1-48. (2004). CIE 1976 uniform colour
    spaces. In CIE 015:2004 Colorimetry, 3rd Edition (p. 24).
    ISBN:978-3-901906-33-6
-   :cite:`Fairchild2010` : Fairchild, M. D., & Wyble, D. R. (2010). hdr-CIELAB
    and hdr-IPT: Simple Models for Describing the Color of High-Dynamic-Range
    and Wide-Color-Gamut Images. Proc. of Color and Imaging Conference,
    322-326. ISBN:978-1-62993-215-6
-   :cite:`Fairchild2011` : Fairchild, M. D., & Chen, P. (2011). Brightness,
    lightness, and specifying color in high-dynamic-range scenes and images. In
    S. P. Farnand & F. Gaykema (Eds.), Proc. SPIE 7867, Image Quality and
    System Performance VIII (p. 78670O). doi:10.1117/12.872075
-   :cite:`Newhall1943a` : Newhall, S. M., Nickerson, D., & Judd, D. B. (1943).
    Final Report of the OSA Subcommittee on the Spacing of the Munsell Colors.
    Journal of the Optical Society of America, 33(7), 385.
    doi:10.1364/JOSA.33.000385
-   :cite:`Wikipedia2001b` : Wikipedia. (2001). Luminance. Retrieved February
    10, 2018, from https://en.wikipedia.org/wiki/Luminance
-   :cite:`Wyszecki2000bd` : Wyszecki, Günther, & Stiles, W. S. (2000). CIE
    1976 (L*u*v*)-Space and Color-Difference Formula. In Color Science:
    Concepts and Methods, Quantitative Data and Formulae (p. 167). Wiley.
    ISBN:978-0-471-39918-6
"""

from __future__ import annotations

import numpy as np

from colour.algebra import spow
from colour.biochemistry import (
    substrate_concentration_MichaelisMenten_Michaelis1913,
    substrate_concentration_MichaelisMenten_Abebe2017,
)
from colour.hints import (
    Any,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Literal,
    Union,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float,
    as_float_array,
    filter_kwargs,
    from_range_1,
    from_range_100,
    get_domain_range_scale,
    to_domain_10,
    to_domain_100,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "luminance_Newhall1943",
    "luminance_ASTMD1535",
    "intermediate_luminance_function_CIE1976",
    "luminance_CIE1976",
    "luminance_Fairchild2010",
    "luminance_Fairchild2011",
    "luminance_Abebe2017",
    "LUMINANCE_METHODS",
    "luminance",
]


def luminance_Newhall1943(V: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *luminance* :math:`R_Y` of given *Munsell* value :math:`V`
    using *Newhall et al. (1943)* method.

    Parameters
    ----------
    V
        *Munsell* value :math:`V`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`R_Y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``R_Y``    | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Newhall1943a`

    Examples
    --------
    >>> luminance_Newhall1943(4.08244375)  # doctest: +ELLIPSIS
    12.5500788...
    """

    V = to_domain_10(V)

    R_Y = (
        1.2219 * V
        - 0.23111 * (V * V)
        + 0.23951 * (V**3)
        - 0.021009 * (V**4)
        + 0.0008404 * (V**5)
    )

    return as_float(from_range_100(R_Y))


def luminance_ASTMD1535(V: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the *luminance* :math:`Y` of given *Munsell* value :math:`V` using
    *ASTM D1535-08e1* method.

    Parameters
    ----------
    V
        *Munsell* value :math:`V`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``V``      | [0, 10]               | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`ASTMInternational2008a`

    Examples
    --------
    >>> luminance_ASTMD1535(4.08244375)  # doctest: +ELLIPSIS
    12.2363426...
    """

    V = to_domain_10(V)

    Y = (
        1.1914 * V
        - 0.22533 * (V**2)
        + 0.23352 * (V**3)
        - 0.020484 * (V**4)
        + 0.00081939 * (V**5)
    )

    return as_float(from_range_100(Y))


def intermediate_luminance_function_CIE1976(
    f_Y_Y_n: FloatingOrArrayLike, Y_n: FloatingOrArrayLike = 100
) -> FloatingOrNDArray:
    """
    Return the *luminance* :math:`Y` in the *luminance* :math:`Y`
    computation for given intermediate value :math:`f(Y/Yn)` using given
    reference white *luminance* :math:`Y_n` as per *CIE 1976* recommendation.

    Parameters
    ----------
    f_Y_Y_n
        Intermediate value :math:`f(Y/Yn)`.
    Y_n
        White reference *luminance* :math:`Y_n`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y`.

    Notes
    -----
    +-------------+-----------------------+---------------+
    | **Domain**  | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``f_Y_Y_n`` | [0, 1]                | [0, 1]        |
    +-------------+-----------------------+---------------+

    +-------------+-----------------------+---------------+
    | **Range**   | **Scale - Reference** | **Scale - 1** |
    +=============+=======================+===============+
    | ``Y``       | [0, 100]              | [0, 100]      |
    +-------------+-----------------------+---------------+

    References
    ----------
    :cite:`CIETC1-482004m`, :cite:`Wyszecki2000bd`

    Examples
    --------
    >>> intermediate_luminance_function_CIE1976(0.495929964178047)
    ... # doctest: +ELLIPSIS
    12.1972253...
    >>> intermediate_luminance_function_CIE1976(0.504482161449319, 95)
    ... # doctest: +ELLIPSIS
    12.1972253...
    """

    f_Y_Y_n = as_float_array(f_Y_Y_n)
    Y_n = as_float_array(Y_n)

    Y = np.where(
        f_Y_Y_n > 24 / 116,
        Y_n * f_Y_Y_n**3,
        Y_n * (f_Y_Y_n - 16 / 116) * (108 / 841),
    )

    return as_float(Y)


def luminance_CIE1976(
    L_star: FloatingOrArrayLike, Y_n: FloatingOrArrayLike = 100
) -> FloatingOrNDArray:
    """
    Return the *luminance* :math:`Y` of given *Lightness* :math:`L^*` with
    given reference white *luminance* :math:`Y_n`.

    Parameters
    ----------
    L_star
        *Lightness* :math:`L^*`
    Y_n
        White reference *luminance* :math:`Y_n`.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L_star`` | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`CIETC1-482004m`, :cite:`Wyszecki2000bd`

    Examples
    --------
    >>> luminance_CIE1976(41.527875844653451)  # doctest: +ELLIPSIS
    12.1972253...
    >>> luminance_CIE1976(41.527875844653451, 95)  # doctest: +ELLIPSIS
    11.5873640...
    """

    L_star = to_domain_100(L_star)
    Y_n = as_float_array(Y_n)

    f_Y_Y_n = (L_star + 16) / 116

    Y = intermediate_luminance_function_CIE1976(f_Y_Y_n, Y_n)

    return as_float(from_range_100(Y))


def luminance_Fairchild2010(
    L_hdr: FloatingOrArrayLike, epsilon: FloatingOrArrayLike = 1.836
) -> FloatingOrNDArray:
    """
    Compute *luminance* :math:`Y` of given *Lightness* :math:`L_{hdr}` using
    *Fairchild and Wyble (2010)* method according to *Michaelis-Menten*
    kinetics.

    Parameters
    ----------
    L_hdr
        *Lightness* :math:`L_{hdr}`.
    epsilon
        :math:`\\epsilon` exponent.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L_hdr``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Fairchild2010`

    Examples
    --------
    >>> luminance_Fairchild2010(31.996390226262736, 1.836)
    ... # doctest: +ELLIPSIS
    0.1219722...
    """

    L_hdr = to_domain_100(L_hdr)

    Y = np.exp(
        np.log(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                L_hdr - 0.02, 100, spow(0.184, epsilon)
            )
        )
        / epsilon
    )

    return as_float(from_range_1(Y))


def luminance_Fairchild2011(
    L_hdr: FloatingOrArrayLike,
    epsilon: FloatingOrArrayLike = 0.474,
    method: Union[Literal["hdr-CIELAB", "hdr-IPT"], str] = "hdr-CIELAB",
) -> FloatingOrNDArray:
    """
    Compute *luminance* :math:`Y` of given *Lightness* :math:`L_{hdr}` using
    *Fairchild and Chen (2011)* method according to *Michaelis-Menten*
    kinetics.

    Parameters
    ----------
    L_hdr
        *Lightness* :math:`L_{hdr}`.
    epsilon
        :math:`\\epsilon` exponent.
    method
        *Lightness* :math:`L_{hdr}` computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L_hdr``  | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Fairchild2011`

    Examples
    --------
    >>> luminance_Fairchild2011(51.852958445912506)  # doctest: +ELLIPSIS
    0.1219722...
    >>> luminance_Fairchild2011(51.643108411718522, method='hdr-IPT')
    ... # doctest: +ELLIPSIS
    0.1219722...
    """

    L_hdr = to_domain_100(L_hdr)
    method = validate_method(method, ["hdr-CIELAB", "hdr-IPT"])

    if method == "hdr-cielab":
        maximum_perception = 247
    else:
        maximum_perception = 246

    Y = np.exp(
        np.log(
            substrate_concentration_MichaelisMenten_Michaelis1913(
                L_hdr - 0.02, maximum_perception, spow(2, epsilon)
            )
        )
        / epsilon
    )

    return as_float(from_range_1(Y))


def luminance_Abebe2017(
    L: FloatingOrArrayLike,
    Y_n: FloatingOrArrayLike = 100,
    method: Union[
        Literal["Michaelis-Menten", "Stevens"], str
    ] = "Michaelis-Menten",
) -> FloatingOrNDArray:
    """
    Compute *luminance* :math:`Y` of *Lightness* :math:`L` using
    *Abebe, Pouli, Larabi and Reinhard (2017)* method according to
    *Michaelis-Menten* kinetics or *Stevens's Power Law*.

    Parameters
    ----------
    L
        *Lightness* :math:`L`.
    Y_n
        Adapting luminance :math:`Y_n` in :math:`cd/m^2`.
    method
        *Luminance* :math:`Y` computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y` in :math:`cd/m^2`.

    Notes
    -----
    -   *Abebe, Pouli, Larabi and Reinhard (2017)* method uses absolute
        luminance levels, thus the domain and range values for the *Reference*
        and *1* scales are only indicative that the data is not affected by
        scale transformations.

    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``L``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+
    | ``Y_n``    | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | ``UN``                | ``UN``        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Abebe2017a`

    Examples
    --------
    >>> luminance_Abebe2017(0.486955571109229)  # doctest: +ELLIPSIS
    12.1972253...
    >>> luminance_Abebe2017(0.474544792145434, method='Stevens')
    ... # doctest: +ELLIPSIS
    12.1972253...
    """

    L = as_float_array(L)
    Y_n = as_float_array(Y_n)
    method = validate_method(method, ["Michaelis-Menten", "Stevens"])

    if method == "stevens":
        Y = np.where(
            Y_n <= 100,
            spow((L + 0.226) / 1.226, 1 / 0.266),
            spow((L + 0.127) / 1.127, 1 / 0.230),
        )
    else:
        Y = np.where(
            Y_n <= 100,
            spow(
                substrate_concentration_MichaelisMenten_Abebe2017(
                    L, 1.448, 0.635, 0.813
                ),
                1 / 0.582,
            ),
            spow(
                substrate_concentration_MichaelisMenten_Abebe2017(
                    L, 1.680, 1.584, 0.096
                ),
                1 / 0.293,
            ),
        )
    Y = Y * Y_n

    return as_float(Y)


LUMINANCE_METHODS: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "Newhall 1943": luminance_Newhall1943,
        "ASTM D1535": luminance_ASTMD1535,
        "CIE 1976": luminance_CIE1976,
        "Fairchild 2010": luminance_Fairchild2010,
        "Fairchild 2011": luminance_Fairchild2011,
        "Abebe 2017": luminance_Abebe2017,
    }
)
LUMINANCE_METHODS.__doc__ = """
Supported *luminance* computation methods.

References
----------
:cite:`ASTMInternational2008a`, :cite:`CIETC1-482004m`, :cite:`Fairchild2010`,
:cite:`Fairchild2011`, :cite:`Newhall1943a`, :cite:`Wyszecki2000bd`

Aliases:

-   'astm2008': 'ASTM D1535'
-   'cie1976': 'CIE 1976'
"""
LUMINANCE_METHODS["astm2008"] = LUMINANCE_METHODS["ASTM D1535"]
LUMINANCE_METHODS["cie1976"] = LUMINANCE_METHODS["CIE 1976"]


def luminance(
    LV: FloatingOrArrayLike,
    method: Union[
        Literal[
            "Abebe 2017",
            "CIE 1976",
            "Glasser 1958",
            "Fairchild 2010",
            "Fairchild 2011",
            "Wyszecki 1963",
        ],
        str,
    ] = "CIE 1976",
    **kwargs: Any,
) -> FloatingOrNDArray:
    """
    Return the *luminance* :math:`Y` of given *Lightness* :math:`L^*` or given
    *Munsell* value :math:`V`.

    Parameters
    ----------
    LV
        *Lightness* :math:`L^*` or *Munsell* value :math:`V`.
    method
        Computation method.

    Other Parameters
    ----------------
    Y_n
        {:func:`colour.colorimetry.luminance_Abebe2017`,
        :func:`colour.colorimetry.luminance_CIE1976`},
        White reference *luminance* :math:`Y_n`.
    epsilon
        {:func:`colour.colorimetry.lightness_Fairchild2010`,
        :func:`colour.colorimetry.lightness_Fairchild2011`},
        :math:`\\epsilon` exponent.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Luminance* :math:`Y`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``LV``     | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``Y``      | [0, 100]              | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Abebe2017a`, :cite:`ASTMInternational2008a`, :cite:`CIETC1-482004m`,
    :cite:`Fairchild2010`, :cite:`Fairchild2011`, :cite:`Newhall1943a`,
    :cite:`Wikipedia2001b`, :cite:`Wyszecki2000bd`

    Examples
    --------
    >>> luminance(41.527875844653451)  # doctest: +ELLIPSIS
    12.1972253...
    >>> luminance(41.527875844653451, Y_n=100)  # doctest: +ELLIPSIS
    12.1972253...
    >>> luminance(42.51993072812094, Y_n=95)  # doctest: +ELLIPSIS
    12.1972253...
    >>> luminance(4.08244375 * 10, method='Newhall 1943')
    ... # doctest: +ELLIPSIS
    12.5500788...
    >>> luminance(4.08244375 * 10, method='ASTM D1535')
    ... # doctest: +ELLIPSIS
    12.2363426...
    >>> luminance(29.829510892279330, epsilon=0.710, method='Fairchild 2011')
    ... # doctest: +ELLIPSIS
    12.1972253...
    """

    LV = as_float_array(LV)
    method = validate_method(method, LUMINANCE_METHODS)

    function = LUMINANCE_METHODS[method]

    # NOTE: "Abebe et al. (2017)" uses absolute luminance levels and has
    # undefined domain-range scale, yet we modify its behaviour consistency
    # with the other methods.
    domain_range_reference = get_domain_range_scale() == "reference"
    domain_range_1 = get_domain_range_scale() == "1"

    domain_1 = (luminance_Fairchild2010, luminance_Fairchild2011)
    domain_10 = (luminance_Newhall1943, luminance_ASTMD1535)
    domain_undefined = (luminance_Abebe2017,)

    if function in domain_10 and domain_range_reference:
        LV = LV / 10

    if function in domain_undefined and domain_range_1:
        LV = LV * 100

    Y_V = function(LV, **filter_kwargs(function, **kwargs))

    if function in domain_1 and domain_range_reference:
        Y_V *= 100

    if function in domain_undefined and domain_range_1:
        Y_V /= 100

    return Y_V



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour.appearance.zcam` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.appearance import (
    VIEWING_CONDITIONS_ZCAM,
    InductionFactors_ZCAM,
    CAM_Specification_ZCAM,
    XYZ_to_ZCAM,
    ZCAM_to_XYZ,
)
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    ignore_numpy_errors,
    tsplit,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_ZCAM",
    "TestZCAM_to_XYZ",
]


class TestXYZ_to_ZCAM(unittest.TestCase):
    """
    Define :func:`colour.appearance.zcam.XYZ_to_ZCAM` definition unit tests
    methods.
    """

    def test_XYZ_to_ZCAM(self):
        """Test :func:`colour.appearance.zcam.XYZ_to_ZCAM` definition."""

        XYZ = np.array([185, 206, 163])
        XYZ_w = np.array([256, 264, 202])
        L_a = 264
        Y_b = 100
        surround = VIEWING_CONDITIONS_ZCAM["Average"]
        np.testing.assert_allclose(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            np.array(
                [
                    92.2520,
                    3.0216,
                    196.3524,
                    19.1314,
                    321.3464,
                    10.5252,
                    237.6401,
                    np.nan,
                    34.7022,
                    25.2994,
                    91.6837,
                ]
            ),
            rtol=0.025,
            atol=0.025,
        )

        XYZ = np.array([89, 96, 120])
        np.testing.assert_allclose(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            np.array(
                [
                    71.2071,
                    6.8539,
                    250.6422,
                    32.7963,
                    248.0394,
                    23.8744,
                    307.0595,
                    np.nan,
                    18.2796,
                    40.4621,
                    70.4026,
                ]
            ),
            rtol=0.025,
            atol=0.025,
        )

        # NOTE: Hue quadrature :math:`H_z` is significantly different for this
        # test, i.e. 47.748252 vs 43.8258.
        # NOTE: :math:`F_L` as reported in the supplemental document has the
        # same value as for :math:`L_a` = 264 instead of 150. The values seem
        # to be computed for :math:`L_a` = 264 and :math:`Y_b` = 100.
        XYZ = np.array([79, 81, 62])
        # L_a = 150
        # Y_b = 60
        surround = VIEWING_CONDITIONS_ZCAM["Dim"]
        np.testing.assert_allclose(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            np.array(
                [
                    68.8890,
                    0.9774,
                    58.7532,
                    12.5916,
                    196.7686,
                    2.7918,
                    43.8258,
                    np.nan,
                    11.0371,
                    44.4143,
                    68.8737,
                ]
            ),
            rtol=0.025,
            atol=4,
        )

        XYZ = np.array([910, 1114, 500])
        XYZ_w = np.array([2103, 2259, 1401])
        L_a = 359
        Y_b = 16
        surround = VIEWING_CONDITIONS_ZCAM["Dark"]
        np.testing.assert_allclose(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            np.array(
                [
                    82.6445,
                    13.0838,
                    123.9464,
                    44.7277,
                    114.7431,
                    18.1655,
                    178.6422,
                    np.nan,
                    34.4874,
                    26.8778,
                    78.2653,
                ]
            ),
            rtol=0.025,
            atol=0.025,
        )

        XYZ = np.array([96, 67, 28])
        np.testing.assert_allclose(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            np.array(
                [
                    33.0139,
                    19.4070,
                    389.7720 % 360,
                    86.1882,
                    45.8363,
                    26.9446,
                    397.3301,
                    np.nan,
                    43.6447,
                    47.9942,
                    30.2593,
                ]
            ),
            rtol=0.025,
            atol=0.025,
        )

    def test_n_dimensional_XYZ_to_ZCAM(self):
        """
        Test :func:`colour.appearance.zcam.XYZ_to_ZCAM` definition
        n-dimensional support.
        """

        XYZ = np.array([185, 206, 163])
        XYZ_w = np.array([256, 264, 202])
        L_a = 264
        Y_b = 100
        surround = VIEWING_CONDITIONS_ZCAM["Average"]
        specification = XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround)

        XYZ = np.tile(XYZ, (6, 1))
        specification = np.tile(specification, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            specification,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            specification,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        specification = np.reshape(specification, (2, 3, 11))
        np.testing.assert_almost_equal(
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround),
            specification,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_domain_range_scale_XYZ_to_ZCAM(self):
        """
        Test :func:`colour.appearance.zcam.XYZ_to_ZCAM` definition
        domain and range scale support.
        """

        XYZ = np.array([185, 206, 163])
        XYZ_w = np.array([256, 264, 202])
        L_a = 264
        Y_b = 100
        surround = VIEWING_CONDITIONS_ZCAM["Average"]
        specification = XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround)

        d_r = (
            ("reference", 1, 1),
            (
                "1",
                1,
                np.array([1, 1, 1 / 360, 1, 1, 1, 1 / 400, np.nan, 1, 1, 1]),
            ),
            (
                "100",
                100,
                np.array(
                    [
                        100,
                        100,
                        100 / 360,
                        100,
                        100,
                        100,
                        100 / 400,
                        np.nan,
                        100,
                        100,
                        100,
                    ]
                ),
            ),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_ZCAM(
                        XYZ * factor_a, XYZ_w * factor_a, L_a, Y_b, surround
                    ),
                    as_float_array(specification) * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_ZCAM(self):
        """
        Test :func:`colour.appearance.zcam.XYZ_to_ZCAM` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_w = np.array(case)
            L_a = case[0]
            Y_b = 100
            surround = InductionFactors_ZCAM(
                case[0], case[0], case[0], case[0]
            )
            XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround)


class TestZCAM_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.appearance.zcam.ZCAM_to_XYZ` definition unit
    tests methods.
    """

    def test_ZCAM_to_XYZ(self):
        """Test :func:`colour.appearance.zcam.ZCAM_to_XYZ` definition."""

        specification = CAM_Specification_ZCAM(
            92.2520,
            3.0216,
            196.3524,
            19.1314,
            321.3464,
            10.5252,
            237.6401,
            np.nan,
            34.7022,
            25.2994,
            91.6837,
        )
        XYZ_w = np.array([256, 264, 202])
        L_a = 264
        Y_b = 100
        surround = VIEWING_CONDITIONS_ZCAM["Average"]
        np.testing.assert_allclose(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            np.array([185, 206, 163]),
            atol=0.01,
            rtol=0.01,
        )

        specification = CAM_Specification_ZCAM(
            71.2071,
            6.8539,
            250.6422,
            32.7963,
            248.0394,
            23.8744,
            307.0595,
            np.nan,
            18.2796,
            40.4621,
            70.4026,
        )
        np.testing.assert_allclose(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            np.array([89, 96, 120]),
            atol=0.01,
            rtol=0.01,
        )

        specification = CAM_Specification_ZCAM(
            68.8890,
            0.9774,
            58.7532,
            12.5916,
            196.7686,
            2.7918,
            43.8258,
            np.nan,
            11.0371,
            44.4143,
            68.8737,
        )
        surround = VIEWING_CONDITIONS_ZCAM["Dim"]
        np.testing.assert_allclose(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            np.array([79, 81, 62]),
            atol=0.01,
            rtol=0.01,
        )

        specification = CAM_Specification_ZCAM(
            82.6445,
            13.0838,
            123.9464,
            44.7277,
            114.7431,
            18.1655,
            178.6422,
            np.nan,
            34.4874,
            26.8778,
            78.2653,
        )
        XYZ_w = np.array([2103, 2259, 1401])
        L_a = 359
        Y_b = 16
        surround = VIEWING_CONDITIONS_ZCAM["Dark"]
        np.testing.assert_allclose(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            np.array([910, 1114, 500]),
            atol=0.01,
            rtol=0.01,
        )

        specification = CAM_Specification_ZCAM(
            33.0139,
            19.4070,
            389.7720 % 360,
            86.1882,
            45.8363,
            26.9446,
            397.3301,
            np.nan,
            43.6447,
            47.9942,
            30.2593,
        )
        np.testing.assert_allclose(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            np.array([96, 67, 28]),
            atol=0.01,
            rtol=0.01,
        )

    def test_n_dimensional_ZCAM_to_XYZ(self):
        """
        Test :func:`colour.appearance.zcam.ZCAM_to_XYZ` definition
        n-dimensional support.
        """

        XYZ = np.array([185, 206, 163])
        XYZ_w = np.array([256, 264, 202])
        L_a = 264
        Y_b = 100
        surround = VIEWING_CONDITIONS_ZCAM["Average"]
        specification = XYZ_to_ZCAM(XYZ, XYZ_w, L_a, Y_b, surround)
        XYZ = ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround)

        specification = CAM_Specification_ZCAM(
            *np.transpose(np.tile(tsplit(specification), (6, 1))).tolist()
        )
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_almost_equal(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            XYZ,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        np.testing.assert_almost_equal(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            XYZ,
            decimal=7,
        )

        specification = CAM_Specification_ZCAM(
            *tsplit(np.reshape(specification, (2, 3, 11))).tolist()
        )
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_almost_equal(
            ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround),
            XYZ,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_domain_range_scale_ZCAM_to_XYZ(self):
        """
        Test :func:`colour.appearance.zcam.ZCAM_to_XYZ` definition
        domain and range scale support.
        """

        XYZ_i = np.array([185, 206, 163])
        XYZ_w = np.array([256, 264, 202])
        L_a = 264
        Y_b = 100
        surround = VIEWING_CONDITIONS_ZCAM["Average"]
        specification = XYZ_to_ZCAM(XYZ_i, XYZ_w, L_a, Y_b, surround)
        XYZ = ZCAM_to_XYZ(specification, XYZ_w, L_a, Y_b, surround)

        d_r = (
            ("reference", 1, 1),
            (
                "1",
                np.array([1, 1, 1 / 360, 1, 1, 1, 1 / 400, np.nan, 1, 1, 1]),
                1,
            ),
            (
                "100",
                np.array(
                    [
                        100,
                        100,
                        100 / 360,
                        100,
                        100,
                        100,
                        100 / 400,
                        np.nan,
                        100,
                        100,
                        100,
                    ]
                ),
                100,
            ),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    ZCAM_to_XYZ(
                        specification * factor_a,
                        XYZ_w * factor_b,
                        L_a,
                        Y_b,
                        surround,
                    ),
                    XYZ * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_raise_exception_ZCAM_to_XYZ(self):
        """
        Test :func:`colour.appearance.zcam.ZCAM_to_XYZ` definition
        raised exception.
        """

        self.assertRaises(
            ValueError,
            ZCAM_to_XYZ,
            CAM_Specification_ZCAM(
                41.731091132513917,
                None,
                219.04843265831178,
            ),
            np.array([256, 264, 202]),
            318.31,
            20.0,
            VIEWING_CONDITIONS_ZCAM["Average"],
        )

    @ignore_numpy_errors
    def test_nan_ZCAM_to_XYZ(self):
        """
        Test :func:`colour.appearance.zcam.ZCAM_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            J = case[0]
            C = case[0]
            h = case[0]
            XYZ_w = np.array(case)
            L_a = case[0]
            Y_b = 100
            surround = InductionFactors_ZCAM(
                case[0], case[0], case[0], case[0]
            )
            ZCAM_to_XYZ(
                CAM_Specification_ZCAM(J, C, h, M=50),
                XYZ_w,
                L_a,
                Y_b,
                surround,
            )



"""Showcases *whiteness* computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"Whiteness" Computations')

XYZ = np.array([95.00000000, 100.00000000, 105.00000000])
XYZ_0 = np.array([94.80966767, 100.00000000, 107.30513595])
message_box(
    f'Computing "whiteness" using "Berger (1959)" method for given sample and '
    f'reference white "CIE XYZ" tristimulus values matrices:\n\n'
    f"\t{XYZ}\n"
    f"\t{XYZ_0}"
)
print(colour.whiteness(XYZ, XYZ_0, method="Berger 1959"))
print(colour.colorimetry.whiteness_Berger1959(XYZ, XYZ_0))

print("\n")

message_box(
    f'Computing "whiteness" using "Taube (1960)" method for given sample and '
    f'reference white "CIE XYZ" tristimulus values matrices:\n\n'
    f"\t{XYZ}\n"
    f"\t{XYZ_0}"
)
print(colour.whiteness(XYZ, XYZ_0, method="Taube 1960"))
print(colour.colorimetry.whiteness_Taube1960(XYZ, XYZ_0))

print("\n")

Lab = colour.XYZ_to_Lab(XYZ / 100, colour.XYZ_to_xy(XYZ_0 / 100))
message_box(
    f'Computing "whiteness" using "Stensby (1968)" method for given sample '
    f'"CIE L*a*b*" colourspace array:\n\n\t{Lab}'
)
print(colour.whiteness(XYZ, XYZ_0, method="Stensby 1968"))
print(colour.colorimetry.whiteness_Stensby1968(Lab))

print("\n")

message_box(
    f'Computing "whiteness" using "ASTM E313" method for given sample '
    f'"CIE XYZ" tristimulus values:\n\n\t{XYZ}'
)
print(colour.whiteness(XYZ, XYZ_0, method="ASTM E313"))
print(colour.colorimetry.whiteness_ASTME313(XYZ))

print("\n")

xy = colour.XYZ_to_xy(XYZ / 100)
Y = 100
message_box(
    f'Computing "whiteness" using "Ganz and Griesser (1979)" method for given '
    f'sample "xy" chromaticity coordinates, "Y" tristimulus value:\n\n'
    f"\t{xy}\n"
    f"\t{Y}"
)
print(colour.whiteness(XYZ, XYZ_0, method="Ganz 1979"))
print(colour.colorimetry.whiteness_Ganz1979(xy, Y))

print("\n")

xy = colour.XYZ_to_xy(XYZ / 100)
Y = 100
xy_n = colour.XYZ_to_xy(XYZ_0 / 100)
message_box(
    f'Computing "whiteness" using "CIE 2004" method for given sample "xy" '
    f'chromaticity coordinates, "Y" tristimulus value and "xy_n" chromaticity '
    f"coordinates of perfect diffuser:\n\n"
    f"\t{xy}\n"
    f"\t{Y}\n"
    f"\t{xy_n}"
)
print(colour.whiteness(XYZ, XYZ_0, xy_n=xy_n))
print(colour.colorimetry.whiteness_CIE2004(xy, Y, xy_n))
